import contextlib as _contextlib
import inspect
import warnings
from collections import OrderedDict

import numpy as np
import pandas as pd


class const:
    TASK_AUTO = 'auto'
    TASK_BINARY = 'binary'
    TASK_MULTICLASS = 'multiclass'
    TASK_REGRESSION = 'regression'
    TASK_MULTILABEL = 'multilabel'


def set_random_state(random_state):
    if random_state is not None:
        seed = random_state if isinstance(random_state, int) \
            else random_state.randint(0, 65535)
        np.random.seed(seed)

        try:
            import torch
            torch.random.manual_seed(seed)
        except:
            pass


def check_cols(data, *x, ):
    x = filter(None, x)
    all_cols = data.columns

    for iter_ in x:
        if isinstance(iter_, str):
            iter_ = {iter_}

        for i in iter_:
            assert i in all_cols, f'Nonexistent variable {i}.'


def to_df(**data):
    dfs = []
    for k, v in data.items():
        if v is None:
            pass  # ignore the item
        elif len(v.shape) == 1:
            dfs.append(pd.Series(v, name=k))
        elif v.shape[1] == 1:
            dfs.append(pd.DataFrame(v, columns=[k]))
        else:
            dfs.append(pd.DataFrame(
                v, columns=[f'{k}_{i}' for i in range(v.shape[1])]))
    df = pd.concat(dfs, axis=1)
    return df


@_contextlib.contextmanager
def context(msg):
    try:
        yield
    except Exception as ex:
        if ex.args:
            msg = u'{}: {}'.format(msg, ex.args[0])
        else:
            msg = str(msg)
        ex.args = (msg,) + ex.args[1:]
        raise


def is_notebook():
    """
    code from https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except:
        return False


def view_pydot(pdot, prog='dot'):
    try:
        from IPython.display import Image, display

        img = Image(pdot.create_png(prog=prog))
        display(img)
    except Exception as e:
        warnings.warn(f'Failed to display pydot image: {e}.')


def unique(y):
    if hasattr(y, 'unique'):
        uniques = set(y.unique())
    else:
        uniques = set(y)
    return uniques


def infer_task_type(y, *, dropna=True, excludes=None, regression_exponent=0.382):
    assert excludes is None or isinstance(excludes, (list, tuple, set))

    if len(y.shape) > 1 and y.shape[-1] > 1:
        labels = list(range(y.shape[-1]))
        task = const.TASK_MULTILABEL
        return task, labels

    uniques = unique(y)
    if dropna:
        if uniques.__contains__(np.nan):
            uniques.remove(np.nan)
        if uniques.__contains__(None):
            uniques.remove(None)
    if excludes is not None and len(excludes) > 0:
        uniques -= set(excludes)
    n_unique = len(uniques)
    labels = []

    if n_unique == 0:
        raise ValueError('Could not infer task type from empty "y"')
    elif n_unique == 1:
        raise ValueError(f'Could not infer task type from unique "{uniques}"')
    elif n_unique == 2:
        task = const.TASK_BINARY
    elif y.dtype.kind == 'f':
        task = const.TASK_REGRESSION
    elif y.dtype.kind == 'i':
        n_sample = len(y)
        if n_unique > n_sample ** regression_exponent:
            task = const.TASK_REGRESSION
        else:
            task = const.TASK_MULTICLASS
    else:
        task = const.TASK_MULTICLASS

    if task != const.TASK_REGRESSION:
        labels = pd.Series(list(uniques)).sort_values().tolist()

    return task, labels


def get_params(obj, include_default=False):
    def _get_init_params(cls):
        init = cls.__init__
        if init is object.__init__:
            return []

        init_signature = inspect.signature(init)
        parameters = [p for p in init_signature.parameters.values()
                      if p.name != 'self']  # and p.kind != p.VAR_KEYWORD]
        return parameters

    fn_get_params = getattr(obj, 'get_params', None)
    if callable(fn_get_params):
        return fn_get_params()

    out = OrderedDict()
    for p in _get_init_params(type(obj)):
        name = p.name
        value = getattr(obj, name, None)
        if include_default or value is not p.default:
            out[name] = value

    return out


def to_repr(obj, excludes=None):
    try:
        if excludes is None:
            excludes = []
        out = ['%s=%r' % (k, v) for k, v in get_params(
            obj).items() if k not in excludes]
        repr_ = ', '.join(out)
        return f'{type(obj).__name__}({repr_})'
    except Exception as e:
        return f'{type(e).__name__}:{e}, at <to_repr>: {type(obj).__name__}'


def to_snake_case(camel_str):
    last_isupper = None
    a = []
    for c in camel_str:
        u = c.isupper()
        if u:
            if last_isupper is False:
                a.append('_')
            a.append(c.lower())
        else:
            a.append(c)
        last_isupper = u
    s = "".join(a)
    while s.startswith("_"):
        s = s[1:]
    return s


def to_camel_case(snake_str):
    components = snake_str.split('_')
    return ''.join(x.title() for x in components)


def drop_none(**kwargs):
    r = {k: v for k, v in kwargs.items() if v is not None}
    return r


def convert2array(data, *S):
    assert isinstance(data, pd.DataFrame)

    def _get_array(cols):
        if cols is not None:
            r = data[cols].values
            if len(r.shape) == 1:
                r = np.expand_dims(r, axis=1)
        else:
            r = None
        return r

    S = map(_get_array, S)

    return tuple(S)



# from . import base_models, double_ml, doubly_robust, meta_learner,\
#     propensity_score

from .approximation_bound import ApproxBound
from .base_models import BaseEstModel
from .causal_tree import CausalTree
from .deepiv import DeepIV
from .double_ml import DML4CATE
from .doubly_robust import DoublyRobust
from .effect_score import RLoss, PredLoss
from .ensemble import EnsembleEstModels
from .iv import NP2SLS
from .meta_learner import SLearner, TLearner, XLearner
from .propensity_score import InversePbWeighting, PropensityScore
from ._permuted import PermutedDoublyRobust
from ._permuted import PermutedSLearner, PermutedTLearner, PermutedXLearner
from ._factory import ESTIMATOR_FACTORIES



"""
Logging utilities, adapted from tf_logging
"""

# pylint: disable=unused-import
# pylint: disable=g-bad-import-order
# pylint: disable=invalid-name
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging as _logging
import os as _os
import re
import sys as _sys
import traceback as _traceback
from logging import DEBUG
from logging import ERROR
from logging import FATAL
from logging import INFO
from logging import WARN

_name2level = {
    'FATAL': FATAL,
    'F': FATAL,

    'ERROR': ERROR,
    'ERR': ERROR,
    'E': ERROR,

    'WARNING': WARN,
    'WARN': WARN,
    'W': WARN,

    'INFO': INFO,
    'I': INFO,

    'DEBUG': DEBUG,
    'D': DEBUG,
}


def _init_log_level():
    pkg = __name__.split('.')[0]
    env_key = pkg.upper() + '_LOG_LEVEL'
    for k, v in _os.environ.items():
        if k.upper() == env_key:
            if v.upper() in _name2level.keys():
                return _name2level[v.upper()]
            elif re.match(r'^\d$', v):
                return int(v)
            else:
                print(f'Unrecognized log level {v}.', file=_sys.stderr)

    _dev_mode = __file__.lower().find('site-packages') < 0
    return INFO if _dev_mode else WARN


# settings
_log_level = _init_log_level()

# _log_format = '%(name)s %(levelname).1s%(asctime)s.%(msecs)d %(filename)s %(lineno)d - %(message)s'
# _log_format = '%(module)s %(levelname).1s %(filename)s %(lineno)d - %(message)s'
# _log_format = '%(levelname).1s %(sname)s.%(filename)s %(lineno)d - %(message)s'
_log_format = '%(asctime)s %(levelname).1s %(sname)s.%(filename)s %(lineno)d - %(message)s'

# _DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
_date_format = '%m-%d %H:%M:%S'


# # detect code run in interactive mode or not
# _interactive = False
# try:
#     # This is only defined in interactive shells.
#     if _sys.ps1: _interactive = True
# except AttributeError:
#     # Even now, we may be in an interactive shell with `python -i`.
#     _interactive = _sys.flags.interactive


class CustomizedLogFormatter(_logging.Formatter):

    def __init__(self, fmt=None, datefmt=None, style='%'):
        super(CustomizedLogFormatter, self).__init__(fmt, datefmt, style)

        self.with_simple_name = fmt.find(style + '(sname)') >= 0

    def formatMessage(self, record):
        if self.with_simple_name:
            record.sname = self.get_simple_name(record.name)

        if isinstance(record.msg, Exception):
            ex = record.msg
            lines = _traceback.format_exception(type(ex), ex, ex.__traceback__)
            record.message = lines[-1] + ''.join(lines[:-1])

        return super(CustomizedLogFormatter, self).formatMessage(record)

    @staticmethod
    def get_simple_name(name):
        if name.endswith('@'):
            return name

        sa = name.split('.')
        if len(sa) <= 1:
            return name

        names = [sa[0]] + \
                [sa[i][0] for i in range(1, len(sa) - 1)]
        return '.'.join(names)


class CustomizedLogger(_logging.Logger):
    FATAL = FATAL
    ERROR = ERROR
    INFO = INFO
    DEBUG = DEBUG
    WARN = WARN

    def __init__(self, name, level=_log_level) -> None:
        super(CustomizedLogger, self).__init__(name, level)

        self.findCaller = _logger_find_caller
        self.setLevel(_log_level)
        self.propagate = False  # disable propagate to parent

        # Don't further configure the logger if the root logger is
        # already configured. This prevents double logging in those cases.
        if not self.handlers:
            # # Add the output handler.
            # stream = _sys.stdout if _interactive else _sys.stderr
            # handler = _logging.StreamHandler(stream)
            # handler.setFormatter(CustomizedLogFormatter(_log_format, _date_format))
            # self.addHandler(handler)

            stdout_handler = _logging.StreamHandler(_sys.stdout)
            stdout_handler.setFormatter(CustomizedLogFormatter(_log_format, _date_format))
            stdout_handler.addFilter(lambda rec: rec.levelno < WARN)
            self.addHandler(stdout_handler)

            stderr_handler = _logging.StreamHandler(_sys.stderr)
            stderr_handler.setFormatter(CustomizedLogFormatter(_log_format, _date_format))
            stderr_handler.addFilter(lambda rec: rec.levelno >= WARN)
            self.addHandler(stderr_handler)

    def getEffectiveLevel(self):
        return _log_level

    def log(self, level, msg, *args, **kwargs):
        super(CustomizedLogger, self).log(level, msg, *args, **kwargs)

    def fatal(self, msg, *args, **kwargs):
        self.log(FATAL, msg, *args, **kwargs)
        self.exception()

    def error(self, msg, *args, **kwargs):
        self.log(ERROR, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.log(WARN, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.log(INFO, msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self.log(DEBUG, msg, *args, **kwargs)

    def log_if(self, level, msg, condition, *args):
        """Log 'msg % args' at level 'level' only if condition is fulfilled."""
        if callable(condition):
            if condition():
                self.log(level, msg, *args)
        elif condition:
            self.log(level, msg, *args)

    def log_every_n(self, level, msg, n, *args):
        """Log 'msg % args' at level 'level' once per 'n' times.

        Logs the 1st call, (N+1)st call, (2N+1)st call,  etc.
        Not threadsafe.

        Args:
          level: The level at which to log.
          msg: The message to be logged.
          n: The number of times this should be called before it is logged.
          *args: The args to be substituted into the msg.
        """
        count = _get_next_log_count_per_token(_get_file_and_line())
        self.log_if(level, msg, not (count % n), *args)

    def log_first_n(self, level, msg, n, *args):  # pylint: disable=g-bad-name
        """Log 'msg % args' at level 'level' only first 'n' times.

        Not threadsafe.

        Args:
          level: The level at which to log.
          msg: The message to be logged.
          n: The number of times this should be called before it is logged.
          *args: The args to be substituted into the msg.
        """
        count = _get_next_log_count_per_token(_get_file_and_line())
        self.log_if(level, msg, count < n, *args)

    def is_debug_enabled(self):
        return self.isEnabledFor(DEBUG)

    def is_info_enabled(self):
        return self.isEnabledFor(INFO)

    def is_warning_enabled(self):
        return self.isEnabledFor(WARN)


def get_logger(name):
    original_logger_class = _logging.getLoggerClass()
    _logging.setLoggerClass(CustomizedLogger)
    logger = _logging.getLogger(name)
    _logging.setLoggerClass(original_logger_class)

    return logger


# compatible with pylog
def getLogger(name):
    return get_logger(name)


def get_level():
    """Return how much logging output will be produced for newer logger."""
    return _log_level


def set_level(v):
    """Sets newer logger threshold for what messages will be logged."""
    assert isinstance(v, (str, int))

    global _log_level

    _log_level = to_level(v)


def to_level(v):
    assert isinstance(v, (int, str)) or v is None

    if v is None:
        v = _log_level
    elif isinstance(v, str):
        if v.upper() in _name2level.keys():
            v = _name2level[v.upper()]
        elif re.match(r'^\d$', v):
            v = int(v)
        else:
            raise ValueError(f'Unrecognized log level {v}.')

    return v


def _get_caller(offset=3):
    """Returns a code and frame object for the lowest non-logging stack frame."""
    # Use sys._getframe().  This avoids creating a traceback object.
    # pylint: disable=protected-access
    f = _sys._getframe(offset)
    # pylint: enable=protected-access
    our_file = f.f_code.co_filename
    f = f.f_back
    while f:
        code = f.f_code
        if code.co_filename != our_file:
            return code, f
        f = f.f_back
    return None, None


# The definition of `findCaller` changed in Python 3.2,
# and further changed in Python 3.8
if _sys.version_info.major >= 3 and _sys.version_info.minor >= 8:

    def _logger_find_caller(stack_info=False, stacklevel=1):  # pylint: disable=g-wrong-blank-lines
        code, frame = _get_caller(4)
        sinfo = None
        if stack_info:
            sinfo = '\n'.join(_traceback.format_stack())
        if code:
            return (code.co_filename, frame.f_lineno, code.co_name, sinfo)
        else:
            return '(unknown file)', 0, '(unknown function)', sinfo
elif _sys.version_info.major >= 3 and _sys.version_info.minor >= 2:

    def _logger_find_caller(stack_info=False):  # pylint: disable=g-wrong-blank-lines
        code, frame = _get_caller(4)
        sinfo = None
        if stack_info:
            sinfo = '\n'.join(_traceback.format_stack())
        if code:
            return (code.co_filename, frame.f_lineno, code.co_name, sinfo)
        else:
            return '(unknown file)', 0, '(unknown function)', sinfo
else:
    def _logger_find_caller():  # pylint: disable=g-wrong-blank-lines
        code, frame = _get_caller(4)
        if code:
            return (code.co_filename, frame.f_lineno, code.co_name)
        else:
            return '(unknown file)', 0, '(unknown function)'

# Counter to keep track of number of log entries per token.
_log_counter_per_token = {}


def _get_next_log_count_per_token(token):
    """Wrapper for _log_counter_per_token.

    Args:
      token: The token for which to look up the count.

    Returns:
      The number of times this function has been called with
      *token* as an argument (starting at 0)
    """
    global _log_counter_per_token  # pylint: disable=global-variable-not-assigned
    _log_counter_per_token[token] = 1 + _log_counter_per_token.get(token, -1)
    return _log_counter_per_token[token]


def _get_file_and_line():
    """Returns (filename, linenumber) for the stack frame."""
    code, f = _get_caller()
    if not code:
        return '<unknown>', 0
    return code.co_filename, f.f_lineno



# TODO: In the current implementation, we believe that the tree model can be a
# valid model for predicting the causal effects estimated by the estimator model. Therefore,
# we can directly apply the fitted tree model to new test dataset and add support
# for such method. In the later version, we may need to verify the correctness of
# doing so and modify the method accordingly.

import numpy as np
import pandas as pd

from sklearn.tree import plot_tree
from sklearn.tree import DecisionTreeRegressor

from ylearn.estimator_model.utils import convert2array


class CEInterpreter:
    """
    Many parameters are similar to those of BaseDecisionTree of sklearn.

    Parameters
    ----------
    criterion : {"squared_error", "friedman_mse", "absolute_error", \
            "poisson"}, default="squared_error"
        The function to measure the quality of a split. Supported criteria
        are "squared_error" for the mean squared error, which is equal to
        variance reduction as feature selection criterion and minimizes the L2
        loss using the mean of each terminal node, "friedman_mse", which uses
        mean squared error with Friedman's improvement score for potential
        splits, "absolute_error" for the mean absolute error, which minimizes
        the L1 loss using the median of each terminal node, and "poisson" which
        uses reduction in Poisson deviance to find splits.        

    splitter : {"best", "random"}, default="best"
        The strategy used to choose the split at each node. Supported
        strategies are "best" to choose the best split and "random" to choose
        the best random split.

    max_depth : int, default=None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node:
        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
        `ceil(min_samples_split * n_samples)` are the minimum
        number of samples for each split.

    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.
            
            - If int, then consider `min_samples_leaf` as the minimum number.
            - If float, then `min_samples_leaf` is a fraction and `ceil(min_samples_leaf * n_samples)` are the minimum number of samples for each node.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : int, float or {"sqrt", "log2"}, default=None
        The number of features to consider when looking for the best split:
        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
        `int(max_features * n_features)` features are considered at each
        split.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.

    random_state : int
        Controls the randomness of the estimator.

    max_leaf_nodes : int, default to None
        Grow a tree with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.
        The weighted impurity decrease equation is the following::
            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)
        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.
        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

    ccp_alpha : non-negative float, default to 0.0
        Value for pruning the tree. #TODO: not implemented yet.
    """
    def __init__(
        self, *,
        criterion='squared_error',
        splitter='best',
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=None,
        random_state=None,
        max_leaf_nodes=None,
        min_impurity_decrease=0.,
        ccp_alpha=0.0,
    ):
        self._is_fitted = False
        self.treatment = None
        self.outcome = None

        self._tree_model = DecisionTreeRegressor(
            criterion=criterion,
            splitter=splitter,
            max_depth=max_depth,
            min_samples_leaf=min_samples_leaf,
            min_samples_split=min_samples_split,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            random_state=random_state,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            ccp_alpha=ccp_alpha
        )

    def fit(
        self,
        data,
        est_model,
        **kwargs
    ):
        #TODO: make this more compatible with that in the policy_interpreter
        """Fit the CEInterpreter model to interpret the causal effect estimated
        by the est_model on data.

        Parameters
        ----------
        data : pandas.DataFrame
            The input samples for the est_model to estimate the causal effects
            and for the CEInterpreter to fit.

        est_model : estimator_model
            est_model should be any valid estimator model of ylearn which was 
            already fitted and can estimate the CATE.
        """
        assert est_model._is_fitted

        self.covariate = est_model.covariate
        assert self.covariate is not None, 'Need covariate to interpret the causal effect.'

        v = convert2array(data, self.covariate)[0]
        n = v.shape[0]
        if hasattr(est_model, 'covariate_transformer'):
            self.cov_transformer = est_model.covariate_transformer
            v = self.cov_transformer.transform(v)

        self._v = v

        causal_effect = est_model.estimate(data=data, quantity=None, **kwargs)

        self._tree_model.fit(v, causal_effect.reshape((n, -1)))

        self._is_fitted = True

        return self

    def interpret(self, *, v=None, data=None):
        """Interpret the fitted model in the test data.

        Parameters
        ----------
        v : numpy.ndarray, optional
            The test covariates, by default None
        data : DataFrame, optional
            The test data in the form of the DataFrame. The model will only use this if v is set as None. In this case, if data is also None, then the data used for trainig will be used, by default None

        Returns
        -------
        dict
            The interpreted results for all examples.
        """
        assert self._is_fitted, 'The model is not fitted yet. Please use the fit method first.'

        v = self._check_features(v=v, data=data)

        node_indicator = self._tree_model.decision_path(v)
        leaf_id = self._tree_model.apply(v)
        feature = self._tree_model.tree_.feature
        threshold = self._tree_model.tree_.threshold

        result_dict = {}
        
        for sample_id, sample in enumerate(v):
            result_dict[f'sample_{sample_id}'] = ''
            node_index = node_indicator.indices[
                node_indicator.indptr[sample_id]: node_indicator.indptr[sample_id + 1]
            ]

            for node_id in node_index:
                if leaf_id[sample_id] == node_id:
                    continue

                if v[sample_id, feature[node_id]] <= threshold[node_id]:
                    thre_sign = '<='
                else:
                    thre_sign = '>'

                feature_id = feature[node_id]
                value = v[sample_id, feature_id]
                thre = threshold[node_id]
                result_dict[f'sample_{sample_id}'] += f'decision node {node_id}: (covariate [{sample_id}, {feature_id}] = {value})' \
                    f' {thre_sign} {thre} \n'
        
        return result_dict

    def _check_features(self, *, v=None, data=None):
        """Validate the training data on predict (probabilities)."""

        if v is not None:
            v = v.reshape(-1, 1) if v.ndim == 1 else v
            assert v.shape[1] == self._tree_model.n_features_in_
            v = v.astype(np.float32)

            return v

        if data is None:
            v = self._v
        else:
            assert isinstance(data, pd.DataFrame)

            v = convert2array(data, self.covariate)[0]
            if hasattr(self, 'cov_transformer'):
                v = self.cov_transformer.transform(v)

            assert v.shape[1] == self._tree_model.n_features_in_
            v = v.reshape(-1, 1) if v.ndim == 1 else v

        v = v.astype(np.float32)

        return v

    def plot(
        self, *,
        feature_names=None,
        max_depth=None,
        class_names=None,
        label='all',
        filled=False,
        node_ids=False,
        proportion=False,
        rounded=False,
        precision=3,
        ax=None,
        fontsize=None
    ):
        """Plot a policy tree.
        The sample counts that are shown are weighted with any sample_weights that
        might be present.
        The visualization is fit automatically to the size of the axis.
        Use the ``figsize`` or ``dpi`` arguments of ``plt.figure``  to control
        the size of the rendering.

        Parameters
        ----------        
        max_depth : int, default=None
            The maximum depth of the representation. If None, the tree is fully
            generated.
        
        class_names : list of str or bool, default=None
            Names of each of the target classes in ascending numerical order.
            Only relevant for classification and not supported for multi-output.
            If ``True``, shows a symbolic representation of the class name.
        
        label : {'all', 'root', 'none'}, default='all'
            Whether to show informative labels for impurity, etc.
            Options include 'all' to show at every node, 'root' to show only at
            the top root node, or 'none' to not show at any node.
        
        filled : bool, default=False
            When set to ``True``, paint nodes to indicate majority class for
            classification, extremity of values for regression, or purity of node
            for multi-output.
        
        impurity : bool, default=True
            When set to ``True``, show the impurity at each node.
        
        node_ids : bool, default=False
            When set to ``True``, show the ID number on each node.
        
        proportion : bool, default=False
            When set to ``True``, change the display of 'values' and/or 'samples'
            to be proportions and percentages respectively.
        
        rounded : bool, default=False
            When set to ``True``, draw node boxes with rounded corners and use
            Helvetica fonts instead of Times-Roman.
        
        precision : int, default=3
            Number of digits of precision for floating point in the values of
            impurity, threshold and value attributes of each node.
        
        ax : matplotlib axis, default=None
            Axes to plot to. If None, use current axis. Any previous content
            is cleared.
        
        fontsize : int, default=None
            Size of text font. If None, determined automatically to fit figure.
        
        Returns
        -------
        annotations : list of artists
            List containing the artists for the annotation boxes making up the
            tree.
        """
        assert self._is_fitted

        impurity = False

        if feature_names == None:
            feature_names = self.covariate

        return plot_tree(
            self._tree_model,
            max_depth=max_depth,
            feature_names=feature_names,
            class_names=class_names,
            label=label,
            filled=filled,
            impurity=impurity,
            node_ids=node_ids,
            proportion=proportion,
            rounded=rounded,
            precision=precision,
            ax=ax,
            fontsize=fontsize,
        )


"""
See (https://arxiv.org/pdf/1803.01422.pdf) for reference.
"""
import math

import numpy as np
import pandas as pd
import torch
import torch.nn as nn

from ylearn.utils import drop_none, set_random_state
from ._base import BaseDiscovery

_is_cuda_available = torch.cuda.is_available()


class L(nn.Module):
    def __init__(self, num_linear, input_features, output_features, dtype=None, device=None):
        super().__init__()

        options = drop_none(dtype=dtype, device=device)

        self.num_linear = num_linear
        self.input_features = input_features
        self.output_features = output_features
        self.weight = nn.Parameter(torch.empty((num_linear, input_features, output_features), **options))
        self.bias = nn.Parameter(torch.empty((num_linear, output_features), **options))
        self.reset_parameters()

    @torch.no_grad()
    def reset_parameters(self):
        k = 1.0 / self.input_features
        bound = math.sqrt(k)
        nn.init.uniform_(self.weight, -bound, bound)
        nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, x):
        out = torch.matmul(x.unsqueeze(dim=2), self.weight.unsqueeze(dim=0))
        out = out.squeeze(dim=2)
        if self.bias is not None:
            out += self.bias
        return out


class A(nn.Module):
    def __init__(self, dim, dtype=None, device=None):
        super(A, self).__init__()
        self.layer_dim = dim
        self.layers = [L(dim[0], dim[i + 1], dim[i + 2], dtype=dtype, device=device) for i in range(len(dim) - 2)]
        self.weights = [layer.weight for layer in self.layers]

    def forward(self, x):
        for layer in self.layers:
            x = layer(torch.sigmoid(x))
        return x


class DagNet(nn.Module):
    def __init__(self, dims, dtype=None, device=None):
        super().__init__()
        d = dims[0]
        self.dims = dims
        self.a1 = nn.Linear(d, d * dims[1], **drop_none(dtype=dtype, device=device))
        self.non_linear = False
        if len(dims) > 2:
            self.non_linear = True
            self.an = A(dims, dtype=dtype, device=device)

    def forward(self, x):
        x = self.a1(x)
        x = x.view(-1, self.dims[0], self.dims[1])
        if self.non_linear:
            x = self.an(x)
        return x.squeeze(dim=2)

    def h_func(self):
        d = self.dims[0]
        a1_weight = self.a1.weight.view(d, -1, d)
        s = torch.sum(a1_weight ** 2, dim=1).t()
        h = torch.trace(torch.matrix_exp(s)) - d
        return h

    def l2_reg(self):
        reg = torch.sum(self.a1.weight ** 2)
        if self.non_linear:
            for ai in self.an.weights:
                reg += torch.sum(ai ** 2)
        return reg

    @torch.no_grad()
    def get_W(self):
        d = self.dims[0]
        a1_weight = self.a1.weight.view(d, -1, d)
        A = torch.sum(a1_weight ** 2, dim=1).t()
        W = torch.sqrt(A)
        W = W.cpu().detach().numpy()
        return W


class CausalDiscovery(BaseDiscovery):
    def __init__(self, hidden_layer_dim=None,
                 lambdaa: float = 0.01, h_tol: float = 1e-6, rho_max: float = 1e6,
                 device=None, random_state=None):
        self.hidden_layer_dim = hidden_layer_dim
        self.lambdaa = lambdaa
        self.h_tol = h_tol
        self.rho_max = rho_max
        self.device = device
        self.random_state = random_state

    @staticmethod
    def _squared_loss(output, target):
        n = target.shape[0]
        loss = 0.5 / n * torch.sum((output - target) ** 2)
        return loss

    def _rho_h_update(self, model, optimizer, X, rho, alpha, h):
        h_new = None
        # X_torch = torch.from_numpy(X)
        # X_torch = torch.tensor(X, dtype=torch.float32, device=self.device)
        while rho < self.rho_max:
            def closure():
                optimizer.zero_grad()
                X_hat = model(X)
                loss = self._squared_loss(X_hat, X)
                h_val = model.h_func()
                penalty = 0.5 * rho * h_val * h_val + alpha * h_val
                l2_reg = 0.5 * self.lambdaa * model.l2_reg()
                primal_obj = loss + penalty + l2_reg
                primal_obj.backward()
                return primal_obj

            optimizer.step(closure)
            with torch.no_grad():
                h_new = model.h_func().item()
            if h_new > 0.5 * h:
                rho *= 10
            else:
                break
        alpha += rho * h_new
        return rho, alpha, h_new

    def __call__(self, data, *, return_dict=False, threshold=None,
                 # optimizer='lbfgs',
                 epoch=100, lr=0.01, max_iter=1500,
                 **kwargs):
        assert isinstance(data, (np.ndarray, pd.DataFrame))

        set_random_state(self.random_state)

        if isinstance(data, pd.DataFrame):
            columns = data.columns.tolist()
            data = data.values
        else:
            columns = None

        device = self.device
        if device is None or device == 'auto':
            device = 'cuda' if _is_cuda_available else 'cpu'
        dtype = torch.float32 if str(device) == 'cuda' else torch.float64
        data_t = torch.tensor(data, dtype=dtype, device=self.device)

        d = data.shape[1]
        hidden = self.hidden_layer_dim if self.hidden_layer_dim is not None else []
        model = DagNet(dims=[d] + hidden + [1], dtype=dtype, device=self.device)
        rho, alpha, h = 1.0, 0.0, np.inf
        for _ in range(epoch):
            optimizer = torch.optim.LBFGS(
                model.parameters(),
                **drop_none(lr=lr, max_iter=max_iter)
            )
            rho, alpha, h = self._rho_h_update(model, optimizer, data_t, rho, alpha, h)
            W_est = model.get_W()
            if h <= self.h_tol or rho >= self.rho_max or np.isnan(W_est).any():
                break
        est = model.get_W()

        if columns is not None:
            est = pd.DataFrame(est, columns=columns, index=columns)

        if return_dict:
            est = self.matrix2dict(est, **drop_none(threshold=threshold))
        elif threshold is not None:
            est[np.abs(est) < threshold] = 0

        return est



from ..policy.policy_model import PolicyTree

from ..utils._common import convert2array


class PolicyInterpreter:
    """
    Attributes
    ----------
    criterion : {'policy_reg'}, default to 'policy_reg' # TODO: may add more criterion

    n_outputs_ : int
        The number of outputs when fit() is performed.

    max_features_ : int
        The inferred value of max_features.

    n_features_in_ : int
        Number of features seen during fit().

    max_depth : int, default to None

    min_samples_split : int or float, default to 2

    min_samples_leaf : int or float, default to 1

    random_state : int

    max_leaf_nodes : int, default to None

    min_impurity_decrease : float, default to 0.0

    ccp_alpha : non-negative float, default to 0.0

    Methods
    ----------
    fit(data, outcome, treatment,
        adjustment=None, covariate=None, treat=None, control=None,)
        Fit the model on data.

    estimate(data=None, quantity=None)
        Estimate the causal effect of the treatment on the outcome in data.

    apply(v)
        Return the index of the leaf that each sample is predicted as.

    decision_path(v)
        Return the decision path.

    _prepare4est(data)
        Prepare for the estimation of the causal effect.

    Reference
    ----------
    This implementation is based on the implementation of BaseDecisionTree
    of sklearn.
    """

    def __init__(
        self, *,
        criterion='policy_reg',
        splitter='best',
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        random_state=2022,
        max_leaf_nodes=None,
        max_features=None,
        min_impurity_decrease=0.0,
        ccp_alpha=0.0,
        min_weight_fraction_leaf=0.0,
    ):
        """
        Many parameters are similar to those of BaseDecisionTree of sklearn.

        Parameters
        ----------
        criterion : {'policy_reg'}, default to 'policy_reg' # TODO: may add more criterion
            The function to measure the quality of a split. The criterion for
            training the tree is (in the Einstein notation)
            
            .. math::
            
                    S = \sum_i g_{ik} y^k_{i},
                    
            where :math:`g_{ik} = \phi(v_i)_k` is a map from the covariates, :math:`v_i`, to a
            basis vector which has only one nonzero element in the :math:`R^k` space. By
            using this criterion, the aim of the model is to find the index of the
            treatment which will render the max causal effect, i.e., finding the
            optimal policy.

        splitter : {"best", "random"}, default="best"
            The strategy used to choose the split at each node. Supported
            strategies are "best" to choose the best split and "random" to choose
            the best random split.

        max_depth : int, default=None
            The maximum depth of the tree. If None, then nodes are expanded until
            all leaves are pure or until all leaves contain less than
            min_samples_split samples.

        min_samples_split : int or float, default=2
            The minimum number of samples required to split an internal node:
            - If int, then consider `min_samples_split` as the minimum number.
            - If float, then `min_samples_split` is a fraction and
            `ceil(min_samples_split * n_samples)` are the minimum
            number of samples for each split.

        min_samples_leaf : int or float, default=1
            The minimum number of samples required to be at a leaf node.
            A split point at any depth will only be considered if it leaves at
            least ``min_samples_leaf`` training samples in each of the left and
            right branches.  This may have the effect of smoothing the model,
            especially in regression.
            - If int, then consider `min_samples_leaf` as the minimum number.
            - If float, then `min_samples_leaf` is a fraction and
            `ceil(min_samples_leaf * n_samples)` are the minimum
            number of samples for each node.

        min_weight_fraction_leaf : float, default=0.0
            The minimum weighted fraction of the sum total of weights (of all
            the input samples) required to be at a leaf node. Samples have
            equal weight when sample_weight is not provided.

        max_features : int, float or {"sqrt", "log2"}, default=None
            The number of features to consider when looking for the best split:
            - If int, then consider `max_features` features at each split.
            - If float, then `max_features` is a fraction and
            `int(max_features * n_features)` features are considered at each
            split.
            - If "sqrt", then `max_features=sqrt(n_features)`.
            - If "log2", then `max_features=log2(n_features)`.
            - If None, then `max_features=n_features`.

        random_state : int
            Controls the randomness of the estimator.

        max_leaf_nodes : int, default to None
            Grow a tree with ``max_leaf_nodes`` in best-first fashion.
            Best nodes are defined as relative reduction in impurity.
            If None then unlimited number of leaf nodes.

        min_impurity_decrease : float, default=0.0
            A node will be split if this split induces a decrease of the impurity
            greater than or equal to this value.
            The weighted impurity decrease equation is the following::
                N_t / N * (impurity - N_t_R / N_t * right_impurity
                                    - N_t_L / N_t * left_impurity)
            where ``N`` is the total number of samples, ``N_t`` is the number of
            samples at the current node, ``N_t_L`` is the number of samples in the
            left child, and ``N_t_R`` is the number of samples in the right child.
            ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
            if ``sample_weight`` is passed.

        ccp_alpha : non-negative float, default to 0.0
            Value for pruning the tree. #TODO: not implemented yet.
        """
        self._is_fitted = False

        self.criterion = criterion
        self.splitter = splitter
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.ccp_alpha = ccp_alpha
        self.random_state = random_state
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features

    def fit(
        self,
        data,
        est_model,
        *,
        covariate=None,
        effect=None,
        effect_array=None,
    ):
        """Fit the PolicyInterpreter model to interpret the policy for the causal
        effect estimated by the est_model on data.

        Parameters
        ----------
        data : pandas.DataFrame
            The input samples for the est_model to estimate the causal effects
            and for the CEInterpreter to fit.

        est_model : estimator_model
            est_model should be any valid estimator model of ylearn which was 
            already fitted and can estimate the CATE.
        """
        covariate = est_model.covariate if covariate is None else covariate
        # outcome = est_model.outcome

        assert covariate is not None, 'Need covariate to interpret the causal effect.'

        if est_model is not None:
            assert est_model._is_fitted
        
        self.covariate = covariate

        # y = convert2array(data, outcome)[0]
        # n, _y_d = y.shape
        # assert _y_d == 1

        self._tree_model = PolicyTree(
            criterion=self.criterion,
            splitter=self.splitter,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            random_state=self.random_state,
            max_leaf_nodes=self.max_leaf_nodes,
            max_features=self.max_features,
            min_impurity_decrease=self.min_impurity_decrease,
            ccp_alpha=self.ccp_alpha,
            min_weight_fraction_leaf=self.min_weight_fraction_leaf,
        )

        self._tree_model.fit(
            data=data,
            covariate=covariate,
            effect=effect,
            effect_array=effect_array,
            est_model=est_model
        )

        self._is_fitted = True

        return self

    def interpret(self, data=None):
        assert self._is_fitted, 'The model is not fitted yet. Please use the fit method first.'

        v = self._tree_model._check_features(v=None, data=data)

        policy_pred = self._tree_model.predict_opt_effect(data=data)
        policy_value = policy_pred.max(1)
        policy_ind = policy_pred.argmax(1)
        node_indicator = self._tree_model.decision_path(v=v)
        leaf_id = self._tree_model.apply(v=v)
        feature = self._tree_model.tree_.feature
        threshold = self._tree_model.tree_.threshold

        result_dict = {}
        
        for sample_id, sample in enumerate(v):
            result_dict[f'sample_{sample_id}'] = ''
            node_index = node_indicator.indices[
                node_indicator.indptr[sample_id]: node_indicator.indptr[sample_id + 1]
            ]

            for node_id in node_index:
                if leaf_id[sample_id] == node_id:
                    continue

                if v[sample_id, feature[node_id]] <= threshold[node_id]:
                    thre_sign = '<='
                else:
                    thre_sign = '>'

                feature_id = feature[node_id]
                value = v[sample_id, feature_id]
                thre = threshold[node_id]
                result_dict[f'sample_{sample_id}'] += f'decision node {node_id}: (covariate [{sample_id}, {feature_id}] = {value})' \
                    f' {thre_sign} {thre} \n'

            result_dict[f'sample_{sample_id}'] += f'The recommended policy is treatment {policy_ind[sample_id]} with value {policy_value[sample_id]}'
            
        return result_dict

    def plot(
        self, *,
        feature_names=None,
        max_depth=None,
        class_names=None,
        label='all',
        filled=False,
        node_ids=False,
        proportion=False,
        rounded=False,
        precision=3,
        ax=None,
        fontsize=None
    ):
        """Plot the tree model.
        The sample counts that are shown are weighted with any sample_weights that
        might be present.
        The visualization is fit automatically to the size of the axis.
        Use the ``figsize`` or ``dpi`` arguments of ``plt.figure``  to control
        the size of the rendering.

        Parameters
        ----------        
        max_depth : int, default=None
            The maximum depth of the representation. If None, the tree is fully
            generated.
        
        class_names : list of str or bool, default=None
            Names of each of the target classes in ascending numerical order.
            Only relevant for classification and not supported for multi-output.
            If ``True``, shows a symbolic representation of the class name.
        
        label : {'all', 'root', 'none'}, default='all'
            Whether to show informative labels for impurity, etc.
            Options include 'all' to show at every node, 'root' to show only at
            the top root node, or 'none' to not show at any node.
        
        filled : bool, default=False
            When set to ``True``, paint nodes to indicate majority class for
            classification, extremity of values for regression, or purity of node
            for multi-output.
        
        impurity : bool, default=True
            When set to ``True``, show the impurity at each node.
        
        node_ids : bool, default=False
            When set to ``True``, show the ID number on each node.
        
        proportion : bool, default=False
            When set to ``True``, change the display of 'values' and/or 'samples'
            to be proportions and percentages respectively.
        
        rounded : bool, default=False
            When set to ``True``, draw node boxes with rounded corners and use
            Helvetica fonts instead of Times-Roman.
        
        precision : int, default=3
            Number of digits of precision for floating point in the values of
            impurity, threshold and value attributes of each node.
        
        ax : matplotlib axis, default=None
            Axes to plot to. If None, use current axis. Any previous content
            is cleared.
        
        fontsize : int, default=None
            Size of text font. If None, determined automatically to fit figure.
        
        Returns
        -------
        annotations : list of artists
            List containing the artists for the annotation boxes making up the
            tree.
        """
        assert self._is_fitted


        if feature_names == None:
            feature_names = self.covariate

        return self._tree_model.plot(
            max_depth=max_depth,
            feature_names=feature_names,
            class_names=class_names,
            label=label,
            filled=filled,
            node_ids=node_ids,
            proportion=proportion,
            rounded=rounded,
            precision=precision,
            ax=ax,
            fontsize=fontsize,
        )


import math
from functools import partial
from itertools import product

import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelEncoder

from ylearn import sklearn_ex as skex
from ylearn.causal_discovery import CausalDiscovery
from ylearn.causal_model import CausalModel, CausalGraph
from ylearn.effect_interpreter.policy_interpreter import PolicyInterpreter
from ylearn.estimator_model import ESTIMATOR_FACTORIES, BaseEstModel
from ylearn.policy.policy_model import PolicyTree
from ylearn.utils import const, logging, view_pydot, infer_task_type, to_repr, drop_none, set_random_state

logger = logging.get_logger(__name__)

DEFAULT_TREATMENT_COUNT_LIMIT_PERCENT = 0.1
DEFAULT_TREATMENT_COUNT_LIMIT_BOUND = 5

GRAPH_STRING_BASE = """
 digraph G {
  graph [splines=true pad=0.5]
  node [shape="box" width=3 height=1.2]

  V [label="Covariates\n\n VLIST"  pos="7.5,2!"]
  X [label="Treatments\n\n XLIST"  pos="5,0!"]
  Y [label="Outcome\n\n YLIST"  pos="10,0!"] 

  V -> {X Y}
  X -> Y
}
"""

GRAPH_STRING_W = """
 digraph G {
  graph [splines=true pad=0.5]
  node [shape="box" width=3 height=1.2]

  W [label="Adjustments\n\n WLIST" pos="5,2!" width=5]
  V [label="Covariates\n\n VLIST"  pos="10,2!"]
  X [label="Treatments\n\n XLIST"  pos="5,0!"]
  Y [label="Outcome\n\n YLIST"  pos="10,0!"] 

  W -> {X Y}
  V -> {X Y}
  X -> Y
}
"""

GRAPH_STRING_Z = """
 digraph G {
  graph [splines=true pad=0.5]
  node [shape="box" width=3 height=1.2]

  V [label="Covariates\n\n VLIST" pos="7,2!" width=5]
  Z [label="Instruments\n\n ZLIST" pos="3,0!"]
  X [label="Treatments\n\n XLIST"  pos="7,0!"  width=2]
  Y [label="Outcome\n\n YLIST"  pos="10,0!" width=2] 

  V -> {X Y}
  Z -> X -> Y
}
"""

GRAPH_STRING_WZ = """
 digraph G {
  graph [splines=true pad=0.5]
  node [shape="box" width=3 height=1.2]

  W [label="Adjustments\n\n WLIST" pos="5,2!" width=5]
  V [label="Covariates\n\n VLIST" pos="10,2!" width=5]
  Z [label="Instruments\n\n ZLIST" pos="3,0!"]
  X [label="Treatments\n\n XLIST"  pos="7,0!"  width=2]
  Y [label="Outcome\n\n YLIST"  pos="10,0!" width=2] 

  W -> {X Y}
  V -> {X Y}
  Z -> X -> Y
}
"""


def _to_list(v, name=None):
    if v is None or isinstance(v, (list, tuple)):
        pass
    elif isinstance(v, str):
        v = [s.strip() for s in v.split(',')]
        v = [s for s in v if len(s) > 0]
    else:
        tag = name if name is not None else 'value'
        raise ValueError(f'Unexpected {tag}: {v}')

    return v


def _join_list(*args):
    r = []
    for a in args:
        if a is None:
            pass
        elif isinstance(a, list):
            r += a
        elif isinstance(a, tuple):
            r += list(a)
        else:
            r += _to_list(a)
    return r


def _empty(v):
    return v is None or len(v) == 0


def _not_empty(v):
    return v is not None and len(v) > 0


def _safe_remove(alist, value, copy=False):
    assert alist is None or isinstance(alist, list)

    if alist is not None:
        if copy:
            alist = alist.copy()

        if isinstance(value, (list, tuple)):
            for v in value:
                if v in alist:
                    alist.remove(v)
        elif value in alist:
            alist.remove(value)

    return alist


def _format(v, line_width=64, line_limit=3):
    if isinstance(v, (list, tuple)):
        lines = []
        line = ''
        for vi in v:
            if len(line) >= line_width:
                if len(lines) + 1 >= line_limit:
                    line += '...'
                    break
                else:
                    lines.append(line)
                    line = ''  # reset new line
            line += f', {vi}' if line else f'{vi}'
        lines.append(line)
        r = ',\n'.join(lines)
    else:
        r = f'{v}'

    return r


def _is_number(dtype):
    return dtype.kind in {'i', 'f'}


def _is_discrete(y):
    return infer_task_type(y)[0] != const.TASK_REGRESSION


def _task_tag(discrete):
    if isinstance(discrete, (np.ndarray, pd.Series)):
        discrete = _is_discrete(discrete)
    return 'classification' if discrete else 'regression'


def _align_task_to_first(data, columns, count_limit):
    """
     select features which has similar task type with the first
    """
    selected = columns[:1]
    discrete = _is_discrete(data[selected[0]])
    for x in columns[1:]:
        x_discrete = _is_discrete(data[x])
        if x_discrete is discrete:
            selected.append(x)
            if len(selected) >= count_limit:
                break

    return selected


def _select_by_task(data, columns, count_limit, discrete):
    """
     select features which has similar task type with the first
    """
    selected = []
    for x in columns:
        x_discrete = _is_discrete(data[x])
        if x_discrete is discrete:
            selected.append(x)
            if len(selected) >= count_limit:
                break

    return selected


class Identifier:
    def identify_treatment(self, data, outcome, discrete_treatment, count_limit, excludes=None):
        raise NotImplemented()

    def identify_aci(self, data, outcome, treatment):
        raise NotImplemented()


class DefaultIdentifier(Identifier):
    def identify_treatment(self, data, outcome, discrete_treatment, count_limit, excludes=None):
        X = data.copy()
        y = X.pop(outcome)

        if excludes is not None and len(excludes) > 0:
            X = X[[c for c in X.columns.tolist() if c not in excludes]]

        tf = skex.FeatureImportancesSelectionTransformer(
            strategy='number', number=X.shape[1], data_clean=False)
        tf.fit(X, y)
        selected = tf.selected_features_

        if discrete_treatment is None:
            treatment = _align_task_to_first(data, selected, count_limit)
        else:
            treatment = _select_by_task(data, selected, count_limit, discrete_treatment)

        return treatment

    def identify_aci(self, data, outcome, treatment):
        adjustment, instrument = None, None

        covariate = [c for c in data.columns.tolist() if c != outcome and c not in treatment]

        return adjustment, covariate, instrument


class IdentifierWithDiscovery(DefaultIdentifier):
    def __init__(self, random_state, **kwargs):
        self.random_state = random_state
        self.discovery_options = kwargs.copy()
        self.causation_matrix_ = None

    def _discovery(self, data, outcome):
        logger.info('discovery causation')

        X = data.copy()
        y = X.pop(outcome)

        if not _is_number(y.dtype):
            y = LabelEncoder().fit_transform(y)

        # preprocessor = skex.general_preprocessor(number_scaler=True)
        preprocessor = skex.general_preprocessor()
        X = preprocessor.fit_transform(X, y)
        X[outcome] = y

        options = dict(random_state=self.random_state)
        if self.discovery_options is not None:
            options.update(self.discovery_options)

        dd = CausalDiscovery(**options)
        return dd(X)

    def identify_treatment(self, data, outcome, discrete_treatment, count_limit, excludes=None):
        causation = self._discovery(data, outcome)
        assert isinstance(causation, pd.DataFrame) and outcome in causation.columns.tolist()

        treatment = causation[outcome].abs().sort_values(ascending=False)
        treatment = [i for i in treatment.index if treatment[i] > 0 and i != outcome]
        if excludes is not None:
            treatment = [t for t in treatment if t not in excludes]

        if len(treatment) > 0:
            if discrete_treatment is None:
                treatment = _align_task_to_first(data, treatment, count_limit)
            else:
                treatment = _select_by_task(data, treatment, count_limit, discrete_treatment)
        else:
            logger.info(f'Not found treatment with causal discovery, so identify treatment by default')
            treatment = super().identify_treatment(data, outcome, discrete_treatment, count_limit, excludes=excludes)

        self.causation_matrix_ = causation

        return treatment

    def identify_aci(self, data, outcome, treatment):
        adjustment, covariate = None, None

        if self.causation_matrix_ is None:
            self.causation_matrix_ = self._discovery(data, outcome)
        causation = self.causation_matrix_
        # threshold = causation.values.diagonal().max()
        threshold = min(np.quantile(causation.values.diagonal(), 0.8),
                        np.mean(causation.values))

        if np.isnan(threshold):
            return super().identify_aci(data, outcome, treatment)

        m = CausalDiscovery().matrix2dict(causation, threshold=threshold)
        cg = CausalGraph(m)
        cm = CausalModel(cg)
        try:
            instrument = cm.get_iv(treatment[0], outcome)
            if _not_empty(instrument):
                instrument = [c for c in instrument if c != outcome and c not in treatment]
            for x in treatment[1:]:
                if _empty(instrument):
                    break
                iv = cm.get_iv(x, outcome)
                if _empty(iv):
                    instrument = None
                    break
                else:
                    iv = [c for c in iv if c != outcome and c not in treatment]
                    instrument = list(set(instrument).intersection(set(iv)))
            if logger.is_info_enabled():
                logger.info(f'found instrument: {instrument}')
        except Exception as e:
            logger.warn(e)
            instrument = []

        if _empty(instrument):
            ids = cm.identify(treatment, outcome, identify_method=('backdoor', 'simple'))
            covariate = list(set(ids['backdoor'][0]))
            covariate = [c for c in covariate if c != outcome and c not in treatment]

        if _empty(covariate):
            logger.info('Not found covariate by discovery, so setup it by default')
            covariate = [c for c in data.columns.tolist()
                         if c != outcome and c not in treatment and c not in instrument]
        else:
            logger.info(f'found covariate: {covariate}')

        if _empty(instrument):
            instrument = None

        return adjustment, covariate, instrument


class Why:
    """
    An all-in-one API for causal learning.

    Parameters
    ----------
    discrete_outcome : bool, default=None
        If True, force the outcome as discrete;
        If False, force the outcome as continuous;
        If None, inferred from outcome.
    discrete_treatment : bool, default=None
        If True, force the treatment variables as discrete;
        If False, force the treatment variables as continuous;
        if None, inferred from the first treatment
    identifier : str, default='auto'
        Available options: 'auto' or 'discovery'
    discovery_model : str, default=None
        Reserved
    discovery_options : dict, default=None
        Parameters (key-values) to initialize the discovery model
    estimator : str, default='auto'
        Name of a valid EstimatorModel. One can also pass an instance of a valid estimator model
    estimator_options : dict, default=None
        Parameters (key-values) to initialize the estimator model.
    random_state : int, default=None
        Random state seed

    Attributes
    ----------
    feature_names_in_ : list of feature names seen during `fit`
    outcome_ : name of outcome
    treatment_ : list of treatment names identified during `fit`
    adjustment_ : list of adjustment names identified during `fit`
    covariate_ : list of covariate names identified during `fit`
    instrument_ : list of instrument names identified during `fit`
    identifier_ : identifier object or None
         Used to identify treatment/adjustment/covariate/instrument if they were not specified during `fit`
    y_encoder_ : LabelEncoder object or None
        Used to encode outcome if its dtype is not numeric
    x_encoders_ : LabelEncoder dict for each treatment if discrete_treatment is True
        Key is the treatment name, value is the LabelEncoder object
        None if discrete_treatment is False
    preprocessor_ : Pipeline object to preprocess data during `fit`
    estimators_ : estimators dict for each treatment
        Key is the treatment name, value is the estimator object

    """

    def __init__(self,
                 discrete_outcome=None,
                 discrete_treatment=None,
                 identifier='auto',
                 discovery_model=None,
                 discovery_options=None,
                 estimator='auto',
                 estimator_options=None,
                 random_state=None):
        assert isinstance(estimator, (str, BaseEstModel))
        if isinstance(estimator, str):
            assert estimator == 'auto' or estimator in ESTIMATOR_FACTORIES.keys()

        self.discrete_outcome = discrete_outcome
        self.discrete_treatment = discrete_treatment
        self.identifier = identifier
        self.discovery_model = discovery_model
        self.discovery_options = discovery_options
        self.estimator = estimator
        self.estimator_options = estimator_options
        self.random_state = random_state

        # fitted
        self._is_fitted = False,
        self.feature_names_in_ = None
        self.treatment_ = None
        self.outcome_ = None
        self.adjustment_ = None
        self.covariate_ = None
        self.instrument_ = None

        #
        self.identifier_ = None
        self.y_encoder_ = None
        self.x_encoders_ = None
        self.preprocessor_ = None
        self.estimators_ = None

    def fit(self, data, outcome,
            *,
            treatment=None,
            adjustment=None,
            covariate=None,
            instrument=None,
            treatment_count_limit=None,
            copy=True,
            **kwargs
            ):
        """
        Fit the Why object, steps:
            1. encode outcome if its dtype is not numeric
            2. identify treatment and adjustment/covariate/instrument
            3. encode treatment if discrete_treatment is True
            4. preprocess data
            5. fit causal estimators

        Parameters
        ----------
        data : pandas.DataFrame, required
        outcome : str, required
            Outcome feature name
        treatment : str or list of str
            Names of the treatment variables.
            If str, will be split into list with comma;
            if None, identified by identifier.
        adjustment : str or list of str, optional
            Names of the adjustment variables. Identified by identifier if adjustment/covariate/instrument are all None.
            If str, will be split into list with comma
        covariate : str or list of str, optional
            Names of the covariate variables. Identified by identifier if adjustment/covariate/instrument are all None.
            If str, will be split into list with comma
        instrument : str or list of str, optional
            Names of the instrument variables. Identified by identifier if adjustment/covariate/instrument are all None.
            If str, will be split into list with comma
        treatment_count_limit : int, default None
            Maximum treatment number, default `min(5, 10% of total feature number)`.
        copy : bool, default True
            Set to False to perform inplace transforming and avoid a copy of data.
        kwargs : options to fit estimators

        Returns
        -------
        instance of :py:class:`Why`
            fitted Why object
        """
        assert isinstance(data, pd.DataFrame)

        feature_names = [c for c in data.columns.tolist()]

        assert outcome is not None and isinstance(outcome, str) and len(outcome) > 0
        assert outcome in feature_names

        set_random_state(self.random_state)
        if copy:
            data = data.copy()

        y = data[outcome]
        if self.discrete_outcome is None:
            self.discrete_outcome = _is_discrete(y)
            logger.info(f'infer outcome as {_task_tag(self.discrete_outcome)}')

        if not _is_number(y.dtype):
            logger.info('encode outcome with LabelEncoder')
            y_encoder = LabelEncoder()
            y = y_encoder.fit_transform(y)
            data[outcome] = y
        else:
            y_encoder = None

        logger.info(f'identify treatment, adjustment, covariate and instrument')
        treatment, adjustment, covariate, instrument = self.identify(
            data, outcome,
            treatment=treatment, adjustment=adjustment, covariate=covariate, instrument=instrument,
            treatment_count_limit=treatment_count_limit,
        )

        # encode treatment
        if self.discrete_treatment:
            logger.info(f'encode treatment ...')
            x_encoders = {}
            for x in treatment:
                x_encoder = LabelEncoder()
                data[x] = x_encoder.fit_transform(data[x])
                x_encoders[x] = x_encoder
        else:
            x_encoders = None

        # preprocess adjustment, covariate, instrument
        logger.info('preprocess data ...')
        preprocessor = skex.general_preprocessor()
        columns = _join_list(adjustment, covariate, instrument)
        assert len(columns) > 0
        data_t = preprocessor.fit_transform(data[columns], y)
        assert isinstance(data_t, pd.DataFrame) and set(data_t.columns.tolist()) == set(columns)
        data[columns] = data_t

        # fit estimator
        estimator_keys = treatment.copy()
        if self.discrete_treatment and len(treatment) > 1:
            estimator_keys.extend(list(filter(lambda _: _[0] != _[1], product(treatment, treatment))))
        estimators = {}
        for x in estimator_keys:
            estimator = self._create_estimator(
                data, outcome, treatment=x,
                adjustment=adjustment, covariate=covariate, instrument=instrument)

            logger.info(f'fit estimator for {x} with {estimator}')
            fit_kwargs = dict(**drop_none(adjustment=adjustment, covariate=covariate, instrument=instrument),
                              **kwargs)
            if isinstance(x, tuple):
                estimator.fit(data, outcome, list(x), **fit_kwargs)
            else:
                estimator.fit(data, outcome, x, **fit_kwargs)
            estimators[x] = estimator

        # save state
        self.feature_names_in_ = feature_names
        self.outcome_ = outcome
        self.treatment_ = treatment
        self.adjustment_ = adjustment
        self.covariate_ = covariate
        self.instrument_ = instrument

        self.estimators_ = estimators

        self.y_encoder_ = y_encoder
        self.x_encoders_ = x_encoders
        self.preprocessor_ = preprocessor
        self._is_fitted = True

        return self

    @staticmethod
    def _get_default_treatment_count_limit(data, outcome):
        return min(math.ceil(data.shape[1] * DEFAULT_TREATMENT_COUNT_LIMIT_PERCENT),
                   DEFAULT_TREATMENT_COUNT_LIMIT_BOUND)

    def _get_default_estimator(self, data, outcome, options, *,
                               treatment, adjustment=None, covariate=None, instrument=None):
        if instrument is not None and len(instrument) > 0:
            estimator = 'iv'
        # elif x_task in {const.TASK_BINARY, const.TASK_MULTICLASS}:
        elif self.discrete_treatment:
            estimator = 'ml'
        else:
            estimator = 'dml'

        return estimator, options

    def identify(self, data, outcome, *,
                 treatment=None, adjustment=None, covariate=None, instrument=None, treatment_count_limit=None):
        """
        Identify treatment and adjustment/covariate/instrument without fitting Why.

        Parameters
        ----------
        data : pandas.DataFrame, required
        outcome : str, outcome feature name, required
        treatment : str or list of str
            Names of the treatment variables.
            If str, will be split into list with comma;
            if None, identified by identifier
        adjustment : str or list of str, optional
            Names of the adjustment variables. Identified by identifier if adjustment/covariate/instrument are all None.
            If str, will be split into list with comma
        covariate : str or list of str, optional
            Names of the covariate variables. Identified by identifier if adjustment/covariate/instrument are all None.
            If str, will be split into list with comma
        instrument : str or list of str, optional
            Names of the instrument variables. Identified by identifier if adjustment/covariate/instrument are all None.
            If str, will be split into list with comma
        treatment_count_limit : int, default None
            Maximum treatment number, default `min(5, 10% of total feature number)`.
        Returns
        -------
        tuple of identified treatment, adjustment, covariate, instrument
        """

        identifier = None

        treatment = _to_list(treatment, name='treatment')
        if _empty(treatment):
            identifier = self._create_identifier()
            if treatment_count_limit is None:
                treatment_count_limit = self._get_default_treatment_count_limit(data, outcome)
            treatment = identifier.identify_treatment(data, outcome, self.discrete_treatment, treatment_count_limit)
            if logger.is_info_enabled():
                tag = 'classification' if _is_discrete(data[treatment[0]]) else 'regression'
                logger.info(f'identified treatment[{tag}]: {treatment}')
        else:
            x_tasks = dict(map(lambda x: (x, 'classification' if _is_discrete(data[x]) else 'regression'),
                               treatment))
            if len(set(x_tasks.values())) > 1:
                logger.warn('Both regression and classification features are used as treatment, '
                            'something maybe unexpected.'
                            f'\n{x_tasks}')

        if self.discrete_treatment is None:
            self.discrete_treatment = _is_discrete(data[treatment[0]])
            if logger.is_info_enabled():
                logger.info(f'infer discrete_treatment={self.discrete_treatment}')

        adjustment = _to_list(adjustment, name='adjustment')
        covariate = _to_list(covariate, name='covariate')
        instrument = _to_list(instrument, name='instrument')

        _safe_remove(adjustment, treatment)
        _safe_remove(covariate, treatment)
        _safe_remove(instrument, treatment)

        if all(map(_empty, (adjustment, covariate, instrument))):
            if identifier is None:
                identifier = self._create_identifier()
            adjustment, covariate, instrument = identifier.identify_aci(data, outcome, treatment)
            if logger.is_info_enabled():
                logger.info(f'identified adjustment: {adjustment}')
                logger.info(f'identified covariate: {covariate}')
                logger.info(f'identified instrument: {instrument}')

        self.identifier_ = identifier
        return treatment, adjustment, covariate, instrument

    def _create_identifier(self):
        if self.identifier == 'discovery':
            options = self.discovery_options if self.discovery_options is not None else {}
            return IdentifierWithDiscovery(self.random_state, **options)
        else:
            return DefaultIdentifier()

    def _create_estimator(self, data, outcome, *,
                          treatment, adjustment=None, covariate=None, instrument=None):
        # x_task, _ = infer_task_type(data[treatment])
        estimator = self.estimator
        options = self.estimator_options if self.estimator_options is not None else {}
        if estimator == 'auto':
            estimator, options = self._get_default_estimator(
                data, outcome, options,
                treatment=treatment,
                adjustment=adjustment,
                covariate=covariate,
                instrument=instrument,
            )

        factory = ESTIMATOR_FACTORIES[estimator](**options)

        estimator = factory(data, outcome, y_task=self.discrete_outcome, x_task=self.discrete_treatment,
                            treatment=treatment,
                            adjustment=adjustment,
                            covariate=covariate,
                            instrument=instrument,
                            random_state=self.random_state)
        return estimator

    def _create_scorers(self, data, scorer):
        if isinstance(scorer, BaseEstModel):
            return {x: scorer for x in self.treatment_}

        if scorer is None or scorer == 'auto':
            scorer = 'rloss'

        factory = ESTIMATOR_FACTORIES[scorer]()
        scorers = {}
        for x in self.treatment_:
            # x_task, _ = infer_task_type(data[x])
            scorer = factory(data, self.outcome_,
                             y_task=self.discrete_outcome,
                             x_task=self.discrete_treatment,
                             treatment=x,
                             adjustment=self.adjustment_,
                             covariate=self.covariate_,
                             instrument=self.instrument_,
                             random_state=self.random_state)
            scorers[x] = scorer
        return scorers

    def _preprocess(self, test_data):
        assert self._is_fitted

        if test_data is not None:
            test_data = test_data.copy()

            if self.preprocessor_ is not None:
                columns = _join_list(self.adjustment_, self.covariate_, self.instrument_)
                assert len(columns) > 0
                test_data[columns] = self.preprocessor_.transform(test_data[columns])

            if self.y_encoder_ is not None and self.outcome_ in test_data.columns.tolist():
                test_data[self.outcome_] = self.y_encoder_.transform(test_data[self.outcome_])

        return test_data

    def _safe_treat_control(self, t_or_c, name):
        assert self._is_fitted

        if t_or_c is None:
            return None

        if isinstance(t_or_c, (tuple, list)):
            assert len(t_or_c) == len(self.treatment_), \
                f'{name} should have the same number with treatment ({self.treatment_})'
        else:
            assert len(self.treatment_) == 1, \
                f'{name} should be list or tuple if the number of treatment is greater than 1'
            t_or_c = [t_or_c, ]

        return t_or_c

    def causal_graph(self):
        """
        Get identified causal graph

        Returns
        -------
        CausalGraph object
        """
        causation = self.identifier_.causation_matrix_ \
            if isinstance(self.identifier_, IdentifierWithDiscovery) else None

        if causation is not None:
            threshold = causation.values.diagonal().max()
            m = CausalDiscovery().matrix2dict(causation, threshold=threshold)
        else:
            m = {}
            fmt = partial(_format, line_limit=1, line_width=20)
            label_y = f'Outcome({self.outcome_})'
            label_x = f'Treatments({fmt(self.treatment_)})'

            if self.adjustment_ is not None and len(self.adjustment_) > 0:
                label_w = f'Adjustments({fmt(self.adjustment_)})'
                m[label_w] = [label_x, label_y]
            if self.covariate_ is not None and len(self.covariate_) > 0:
                label_v = f'Covariates({fmt(self.covariate_)})'
                m[label_v] = [label_x, label_y]
            if self.instrument_ is not None and len(self.instrument_) > 0:
                label_z = f'Instruments({fmt(self.instrument_)})'
                m[label_z] = [label_x, ]

        cg = CausalGraph(m)
        return cg

    def causal_effect(self, test_data=None, treat=None, control=None, return_detail=False):
        """
        Estimate the causal effect.

        Parameters
        ----------
        test_data : pd.DataFrame, default None
            The test data to evaluate the causal effect.
            If None, the training data is used.
        treat : treatment value or list or ndarray or pandas.Series, default None
            In the case of single discrete treatment, treat should be an int or
            str of one of all possible treatment values which indicates the
            value of the intended treatment;
            in the case of multiple discrete treatment, treat should be a list
            where treat[i] indicates the value of the i-th intended treatment,
            for example, when there are multiple discrete treatments,
            list(['run', 'read']) means the treat value of the first treatment is taken as 'run'
            and that of the second treatment is taken as 'read';
            in the case of continuous treatment, treat should be a float or a ndarray or pandas.Series,
            by default None
        control : treatment value or list or ndarray or pandas.Series, default None
            This is similar to the cases of treat, by default None
        return_detail: bool, default False
            If True, return effect details in result.

        Returns
        -------
        pd.DataFrame
            causal effect of each treatment. The result DataFrame columns are:
                mean: mean of causal effect,
                min: minimum of causal effect,
                max: maximum of causal effect,
                detail(if return_detail is True ): causal effect ndarray;
            In the case of discrete treatment, the result DataFrame indices are multiindex of
            (treatment name and treat_vs_control);
            in the case of continuous treatment, the result DataFrame indices are treatment names.

        """
        test_data = self._preprocess(test_data)
        treat = self._safe_treat_control(treat, 'treat')
        control = self._safe_treat_control(control, 'control')

        if self.discrete_treatment:
            return self._causal_effect_discrete(test_data, treat, control, return_detail=return_detail)
        else:
            return self._causal_effect_continuous(test_data, treat, control, return_detail=return_detail)

    def _causal_effect_discrete(self, test_data=None, treat=None, control=None, return_detail=False):
        dfs = []
        for i, x in enumerate(self.treatment_):
            est = self.estimators_[x]
            xe = self.x_encoders_[x]
            if control is not None:
                assert control[i] in xe.classes_.tolist(), f'Invalid {x} control "{control[i]}"'
                control_i = control[i]
            else:
                control_i = xe.classes_[0]

            if treat is not None:
                assert treat[i] in xe.classes_.tolist(), f'Invalid {x} treat "{treat[i]}"'
                treats = [treat[i], ]
            else:
                if test_data is None:
                    treats = xe.classes_.tolist()
                else:
                    treats = np.unique(test_data[x]).tolist()
                treats = filter(lambda _: _ != control_i, treats)

            if test_data is not None:
                test_data[x] = xe.transform(test_data[x])
            c = xe.transform([control_i])[0]
            for treat_i in treats:
                t = xe.transform([treat_i])[0]
                effect = est.estimate(data=test_data, treat=t, control=c)
                s = pd.Series(dict(mean=effect.mean(),
                                   min=effect.min(),
                                   max=effect.max(),
                                   std=effect.std(),
                                   ),
                              name=(x, f'{treat_i} vs {control_i}')
                              )
                if return_detail:
                    s['detail'] = effect.ravel()
                dfs.append(s)

        result = pd.concat(dfs, axis=1).T
        return result

    def _causal_effect_continuous(self, test_data=None, treat=None, control=None, return_detail=False):
        dfs = []
        for i, x in enumerate(self.treatment_):
            est = self.estimators_[x]
            treat_i = treat[i] if treat is not None else None
            control_i = control[i] if control is not None else None
            if isinstance(treat_i, (pd.Series, pd.DataFrame)):
                treat_i = treat_i.values
            if isinstance(control_i, (pd.Series, pd.DataFrame)):
                control_i = control_i.values
            if treat_i is not None and control_i is None:
                control_i = np.zeros_like(treat_i)
            elif treat_i is None and control_i is not None:
                treat_i = np.ones_like(control_i)
            effect = est.estimate(data=test_data, treat=treat_i, control=control_i)
            s = pd.Series(dict(mean=effect.mean(),
                               min=effect.min(),
                               max=effect.max(),
                               std=effect.std()),
                          name=x)
            if return_detail:
                s['detail'] = effect.ravel()
            dfs.append(s)
        return pd.concat(dfs, axis=1).T

    def individual_causal_effect(self, test_data, control=None):
        """
        Estimate the causal effect for each individual.

        Parameters
        ----------
        test_data : pd.DataFrame, required
            The test data to evaluate the causal effect.
        control : treatment value or list or ndarray or pandas.Series, default None
            In the case of single discrete treatment, control should be an int or
            str of one of all possible treatment values which indicates the
            value of the intended treatment;
            in the case of multiple discrete treatment, control should be a list
            where control[i] indicates the value of the i-th intended treatment,
            for example, when there are multiple discrete treatments,
            list(['run', 'read']) means the treat value of the first treatment is taken as 'run'
            and that of the second treatment is taken as 'read';
            in the case of continuous treatment, treat should be a float or a ndarray or pandas.Series,
            by default None

        Returns
        -------
        pd.DataFrame
            individual causal effect of each treatment.
            The result DataFrame columns are the treatment names;
            In the case of discrete treatment, the result DataFrame indices are multiindex of
            (individual index in test_data, treatment name and treat_vs_control);
            in the case of continuous treatment, the result DataFrame indices are multiindex of
            (individual index in test_data, treatment name).
        """
        assert test_data is not None

        test_data = self._preprocess(test_data)
        control = self._safe_treat_control(control, 'control')

        if self.discrete_treatment:
            return self._individual_causal_effect_discrete(test_data, control)
        else:
            return self._individual_causal_effect_continuous(test_data, control)

    def _individual_causal_effect_discrete(self, test_data, control=None):
        if control is None:
            control = [self.x_encoders_[x].classes_.tolist()[0] for x in self.treatment_]

        dfs = []
        for ri in test_data.index:
            row_df = test_data.loc[ri:ri]
            for xi, x in enumerate(self.treatment_):
                treat_i = row_df[x].tolist()[0]
                control_i = control[xi]
                if treat_i == control_i:
                    effect = 0.0
                else:
                    xe = self.x_encoders_[x]
                    row_df = row_df.copy()
                    row_df[x] = xe.transform(row_df[x])
                    t, c = xe.transform([treat_i, control_i]).tolist()
                    est = self.estimators_[x]
                    effect = est.estimate(data=row_df, treat=t, control=c).ravel()[0]
                s = pd.Series(dict(effect=effect),
                              name=(ri, x, f'{treat_i} vs {control_i}'))
                dfs.append(s)
        return pd.concat(dfs, axis=1).T

    def _individual_causal_effect_continuous(self, test_data, control=None):
        dfs = []
        for i, x in enumerate(self.treatment_):
            est = self.estimators_[x]
            treat_i = test_data[x]
            control_i = control[i] if control is not None else None
            effect = est.estimate(data=test_data, treat=treat_i, control=control_i)
            s = pd.Series(effect.ravel(), name=x)
            dfs.append(s)
        return pd.concat(dfs, axis=1)

    def whatif(self, test_data, new_value, treatment=None):
        """
        Get counterfactual predictions when treatment is changed to new_value from its observational counterpart.

        Parameters
        ----------
        test_data : pd.DataFrame, required
            The test data to predict.
        new_value : ndarray or pd.Series, required
            It should have the same length with test_data.
        treatment : str, default None
            Treatment name.
            If str, it should be on of the fitted attribute **treatment_**.
            If None, then first element in the attribute **treatment_** is used.
        Returns
        -------
        pd.Series
            The counterfactual prediction
        """
        assert test_data is not None and new_value is not None
        assert treatment is None or isinstance(treatment, str)
        if isinstance(treatment, str):
            assert treatment in self.treatment_
        if treatment is None:
            treatment = self.treatment_[0]

        data_t = self._preprocess(test_data)
        estimator = self.estimators_[treatment]
        if estimator.is_discrete_treatment:
            return self._whatif_discrete(data_t, new_value, treatment, estimator)
        else:
            return self._whatif_continuous(data_t, new_value, treatment, estimator)

    def _whatif_discrete(self, data, new_value, treatment, estimator):
        y_old = data[self.outcome_]
        old_value = data[treatment]
        xe = self.x_encoders_[treatment]

        for x in self.treatment_:
            data[x] = self.x_encoders_[x].transform(data[x])

        df = pd.DataFrame(dict(c=old_value, t=new_value), index=old_value.index)
        df['tc'] = df[['t', 'c']].apply(tuple, axis=1)
        tc_pairs = df['tc'].unique().tolist()
        effect = []
        for t, c in tc_pairs:
            tc_rows = df[df['tc'] == (t, c)]
            if t == c:
                eff = np.zeros((len(tc_rows),))
            else:
                data_rows = data.loc[tc_rows.index]
                t_encoded = xe.transform([t])[0]
                c_encoded = xe.transform([c])[0]
                eff = estimator.estimate(data_rows, treat=t_encoded, control=c_encoded)
            effect.append(pd.DataFrame(dict(e=eff.ravel()), index=tc_rows.index))
        effect = pd.concat(effect, axis=0)
        assert len(effect) == len(df)
        df['e'] = effect['e']  # align indices

        y_new = y_old + df['e']
        return y_new

    def _whatif_continuous(self, data, new_value, treatment, estimator):
        y_old = data[self.outcome_]
        old_value = data[treatment].values
        if isinstance(new_value, pd.Series):
            new_value = new_value.values
        effect = estimator.estimate(data, treat=new_value, control=old_value)
        y_new = y_old + effect.ravel()
        return y_new

    def score(self, test_data=None, treat=None, control=None, scorer='auto'):
        """
        Scoring the fitted estimator models.

        Parameters
        ----------
        test_data : pd.DataFrame, required
        treat : treatment value or list or ndarray or pandas.Series, default None
            In the case of single discrete treatment, treat should be an int or
            str of one of all possible treatment values which indicates the
            value of the intended treatment;
            in the case of multiple discrete treatment, treat should be a list
            where treat[i] indicates the value of the i-th intended treatment,
            for example, when there are multiple discrete treatments,
            list(['run', 'read']) means the treat value of the first treatment is taken as 'run'
            and that of the second treatment is taken as 'read';
            in the case of continuous treatment, treat should be a float or a ndarray or pandas.Series,
            by default None
        control : int or list, default None
            This is similar to the cases of treat, by default None
        scorer: str, default 'auto'
            Reserved.

        Returns
        -------
        float
            score
        """
        assert test_data is not None

        scorers = self._create_scorers(test_data, scorer=scorer)
        fit_options = drop_none(adjustment=self.adjustment_, covariate=self.covariate_, instrument=self.instrument_)
        test_data = self._preprocess(test_data)
        treat = self._safe_treat_control(treat, 'treat')
        control = self._safe_treat_control(control, 'control')

        sa = []
        for i, (x, scorer) in enumerate(scorers.items()):
            logger.info(f'fit scorer for {x} with {scorer}')
            if self.discrete_treatment:
                xe = self.x_encoders_[x]
                test_data[x] = xe.transform(test_data[x])
                treat_i = xe.transform([treat[i], ])[0] if treat is not None else None
                control_i = xe.transform([control[i], ])[0] if control is not None else None
            else:
                treat_i = treat[i] if treat is not None else None
                control_i = control[i] if control is not None else None
            scorer.fit(test_data, self.outcome_, x, **fit_options)
            estimator = self.estimators_[x]
            score_options = drop_none(treat=treat_i, control=control_i)
            sa.append(scorer.score(estimator, **score_options))
        score = np.mean(sa)

        return score

    def _effect_array(self, preprocessed_data, treatment, control=None):
        if treatment is None:
            treatment = self.treatment_[:2]
        else:
            treatment = _to_list(treatment, 'treatment')

        if len(treatment) > 2:
            raise ValueError(f'2 treatment are supported at most.')

        if self.discrete_treatment:
            for x in self.treatment_:
                preprocessed_data[x] = self.x_encoders_[x].transform(preprocessed_data[x])

        if self.discrete_treatment and len(treatment) > 1:
            estimator = self.estimators_[tuple(treatment)]
            if control is not None:
                control = [self.x_encoders_[x].transform([control[i]])[0] for i, x in enumerate(treatment)]
            effect = estimator.effect_nji(preprocessed_data, **drop_none(control=control))
            effect_array = effect.reshape(-1, effect.shape[2])
        else:
            effects = []
            for i, x in enumerate(treatment):
                estimator = self.estimators_[x]
                if self.discrete_treatment:
                    xe = self.x_encoders_[x]
                    ci = xe.transform([control[i], ])[0] if control is not None else None
                else:
                    ci = control[i] if control is not None else None
                effect = estimator.effect_nji(preprocessed_data, **drop_none(control=ci))
                assert isinstance(effect, np.ndarray)
                assert effect.ndim == 3 and effect.shape[1] == 1
                effects.append(effect.reshape(-1, effect.shape[2]))
            effect_array = np.hstack(effects)
        return effect_array

    def policy_tree(self, test_data, treatment=None, control=None, **kwargs):
        """
        Get the policy tree

        Parameters
        ----------
        test_data : pd.DataFrame, required
        treatment: str or list, optional
            Treatment names, should be one or two element.
            default the first two elements in attribute **treatment_**
        control : treatment value or list or ndarray or pandas.Series, default None
            In the case of single discrete treatment, control should be an int or
            str of one of all possible treatment values which indicates the
            value of the intended treatment;
            in the case of multiple discrete treatment, control should be a list
            where control[i] indicates the value of the i-th intended treatment,
            for example, when there are multiple discrete treatments,
            list(['run', 'read']) means the control value of the first treatment is taken as 'run'
            and that of the second treatment is taken as 'read';
            in the case of continuous treatment, control should be a float or a ndarray or pandas.Series,
            by default None
        kwargs : dict
            options to initialize the PolicyTree

        Returns
        -------
        PolicyTree :
            The fitted PolicyTree object
        """
        test_data = self._preprocess(test_data)
        control = self._safe_treat_control(control, 'control')
        effect_array = self._effect_array(test_data, treatment, control=control)
        ptree = PolicyTree(**kwargs)
        ptree.fit(test_data, covariate=self.covariate_, effect_array=effect_array)

        return ptree

    def policy_interpreter(self, test_data, treatment=None, control=None, **kwargs):
        """
        Get the policy interpreter

        Parameters
        ----------
        test_data : pd.DataFrame, required
        treatment: treatment names, optional
            Should be one or two element.
            default the first two element in attribute **treatment_**
        control : treatment value or list or ndarray or pandas.Series, default None
            In the case of single discrete treatment, treat should be an int or
            str of one of all possible treatment values which indicates the
            value of the intended treatment;
            in the case of multiple discrete treatment, treat should be a list
            where treat[i] indicates the value of the i-th intended treatment,
            for example, when there are multiple discrete treatments,
            list(['run', 'read']) means the treat value of the first treatment is taken as 'run'
            and that of the second treatment is taken as 'read';
            in the case of continuous treatment, treat should be a float or a ndarray or pandas.Series,
            by default None
        kwargs : options to initialize the PolicyInterpreter

        Returns
        -------
        PolicyInterpreter :
            The fitted PolicyInterpreter object
        """
        test_data = self._preprocess(test_data)
        control = self._safe_treat_control(control, 'control')
        effect_array = self._effect_array(test_data, treatment, control=control)
        pi = PolicyInterpreter(**kwargs)
        pi.fit(test_data, covariate=self.covariate_, est_model=None, effect_array=effect_array)
        return pi

    def plot_policy_tree(self, test_data, treatment=None, control=None, **kwargs):
        """
        Plot the policy tree

        Returns
        -------
        PolicyTree :
            The fitted PolicyTree object
        """
        ptree = self.policy_tree(test_data, treatment=treatment, control=control, **kwargs)
        ptree.plot()
        return ptree

    def plot_policy_interpreter(self, test_data, treatment=None, control=None, **kwargs):
        """
        Plot the policy interpreter

        Returns
        -------
        PolicyInterpreter :
            The fitted PolicyInterpreter object
        """
        pi = self.policy_interpreter(test_data, treatment=treatment, control=control, **kwargs)
        pi.plot()
        return pi

    def plot_causal_graph(self):
        """
        Plot the causal graph.
        """
        import pydot

        values = dict(WLIST=self.adjustment_, VLIST=self.covariate_, XLIST=self.treatment_,
                      YLIST=self.outcome_, ZLIST=self.instrument_)
        if not _empty(self.instrument_) and not _empty(self.adjustment_):
            dot_string = GRAPH_STRING_WZ
        elif not _empty(self.instrument_):
            dot_string = GRAPH_STRING_Z
        elif not _empty(self.adjustment_):
            dot_string = GRAPH_STRING_W
        else:
            dot_string = GRAPH_STRING_BASE

        for k, v in values.items():
            if dot_string.find(k) >= 0:
                width = 40 if k == 'ZLIST' else 64
                dot_string = dot_string.replace(k, _format(v, line_width=width))
        graph = pydot.graph_from_dot_data(dot_string)[0]
        view_pydot(graph, prog='fdp')

    def __repr__(self):
        return to_repr(self)



import pandas as pd

from ylearn import sklearn_ex as skex
from ylearn.utils import const, logging, to_repr, to_snake_case

logger = logging.get_logger(__name__)

ESTIMATOR_FACTORIES = {}


def register(name=None):
    def wrap(cls):
        assert issubclass(cls, BaseEstimatorFactory)

        if name is not None:
            tag = name
        else:
            tag = cls.__name__
            if tag.endswith('Factory'):
                tag = tag[:-7]
            tag = to_snake_case(tag)

        assert tag not in ESTIMATOR_FACTORIES.keys()

        ESTIMATOR_FACTORIES[tag] = cls
        return cls

    return wrap


class BaseEstimatorFactory:
    def __call__(self, data, outcome, treatment, y_task, x_task,
                 adjustment=None, covariate=None, instrument=None, random_state=None):
        raise NotImplemented()

    @staticmethod
    def _model(data, estimator, task, random_state=None, **kwargs):
        assert task is not None
        assert estimator is not None

        return skex.general_estimator(data, task=task, estimator=estimator, random_state=random_state, **kwargs)

    @staticmethod
    def _cf_fold(data):
        size = data.shape[0]
        if size < 3000:
            return 1
        elif size < 10000:
            return 3
        else:
            return 5

    def __repr__(self):
        return to_repr(self)


@register()
class DMLFactory(BaseEstimatorFactory):
    def __init__(self, y_model='rf', x_model='rf', yx_model='lr'):
        self.y_model = y_model
        self.x_model = x_model
        self.yx_model = yx_model

    def __call__(self, data, outcome, treatment, y_task, x_task,
                 adjustment=None, covariate=None, instrument=None, random_state=None):
        from ylearn.estimator_model.double_ml import DML4CATE
        # assert adjustment is not None
        assert covariate is not None

        return DML4CATE(
            y_model=self._model(data, task=y_task, estimator=self.y_model, random_state=random_state),
            x_model=self._model(data, task=x_task, estimator=self.x_model, random_state=random_state),
            yx_model=self._model(data, task=const.TASK_REGRESSION, estimator=self.yx_model, random_state=random_state),
            is_discrete_treatment=x_task if isinstance(x_task, bool) else x_task != const.TASK_REGRESSION,
            cf_fold=self._cf_fold(data),
            random_state=random_state,
        )


@register()
class DRFactory(BaseEstimatorFactory):
    def __init__(self, y_model='gb', x_model='rf', yx_model='gb'):
        self.y_model = y_model
        self.x_model = x_model
        self.yx_model = yx_model

    def __call__(self, data, outcome, treatment, y_task, x_task,
                 adjustment=None, covariate=None, instrument=None, random_state=None):
        from ylearn.estimator_model import PermutedDoublyRobust

        # assert adjustment is not None
        assert x_task != const.TASK_REGRESSION, 'DoublyRobust support discrete treatment only.'

        return PermutedDoublyRobust(
            y_model=self._model(data, task=y_task, estimator=self.y_model, random_state=random_state),
            x_model=self._model(data, task=x_task, estimator=self.x_model, random_state=random_state),
            yx_model=self._model(data, task=const.TASK_REGRESSION, estimator=self.yx_model, random_state=random_state),
            cf_fold=self._cf_fold(data),
            random_state=random_state,
        )


@register()
@register(name='ml')
class MetaLeanerFactory(BaseEstimatorFactory):
    def __init__(self, leaner='tleaner', model='gb'):
        assert leaner.strip().lower()[0] in {'s', 't', 'x'}

        self.leaner = leaner
        self.model = model

    def __call__(self, data, outcome, treatment, y_task, x_task,
                 adjustment=None, covariate=None, instrument=None, random_state=None):
        from ylearn.estimator_model import PermutedSLearner, PermutedTLearner, PermutedXLearner

        # assert adjustment is not None
        assert x_task != const.TASK_REGRESSION, 'MetaLearner support discrete treatment only.'

        tag = self.leaner.strip().lower()[0]
        learners = dict(s=PermutedSLearner, t=PermutedTLearner, x=PermutedXLearner)
        est_cls = learners[tag]
        return est_cls(
            model=self._model(data, task=y_task, estimator=self.model, random_state=random_state),
            is_discrete_outcome=y_task if isinstance(y_task, bool) else y_task != const.TASK_REGRESSION,
            is_discrete_treatment=x_task if isinstance(x_task, bool) else x_task != const.TASK_REGRESSION,
            random_state=random_state,
            # combined_treatment=False,
        )


@register()
@register(name='tree')
class CausalTreeFactory(BaseEstimatorFactory):
    # def __init__(self):
    #     pass
    def __call__(self, data, outcome, treatment, y_task, x_task,
                 adjustment=None, covariate=None, instrument=None, random_state=None):
        from ylearn.estimator_model.causal_tree import CausalTree

        assert adjustment is not None

        return CausalTree(random_state=random_state)  # FIXME


@register()
@register(name='bound')
class ApproxBoundFactory(BaseEstimatorFactory):
    def __init__(self, y_model='gb', x_model='rf', x_prob=None):
        self.y_model = y_model
        self.x_model = x_model
        self.x_prob = x_prob

    def __call__(self, data, outcome, treatment, y_task, x_task,
                 adjustment=None, covariate=None, instrument=None, random_state=None):
        from ylearn.estimator_model.approximation_bound import ApproxBound

        assert covariate is not None

        return ApproxBound(
            y_model=self._model(data, task=y_task, estimator=self.y_model, random_state=random_state),
            x_model=self._model(data, task=x_task, estimator=self.x_model, random_state=random_state),
            x_prob=self.x_prob,
            is_discrete_treatment=x_task if isinstance(x_task, bool) else x_task != const.TASK_REGRESSION,
            random_state=random_state,
        )


@register()
class IVFactory(BaseEstimatorFactory):
    def __init__(self, y_model='lr', x_model='rf'):
        self.y_model = y_model
        self.x_model = x_model

    def __call__(self, data, outcome, treatment, y_task, x_task,
                 adjustment=None, covariate=None, instrument=None, random_state=None):
        from ylearn.estimator_model.iv import NP2SLS

        assert instrument is not None

        return NP2SLS(
            y_model=self._model(data, task=y_task, estimator=self.y_model, random_state=random_state),
            x_model=self._model(data, task=x_task, estimator=self.x_model, random_state=random_state),
            is_discrete_outcome=y_task if isinstance(y_task, bool) else y_task != const.TASK_REGRESSION,
            is_discrete_treatment=x_task if isinstance(x_task, bool) else x_task != const.TASK_REGRESSION,
            random_state=random_state,
        )


try:
    import torch
    from ylearn.estimator_model.deepiv import DeepIV


    class DeepIVWrapper(DeepIV):
        def fit(self, data, outcome, treatment, **kwargs):
            data = self._f64to32(data)
            return super().fit(data, outcome, treatment=treatment, **kwargs)

        def estimate(self, data=None, *args, **kwargs, ):
            if data is not None:
                data = self._f64to32(data)

            effect = super().estimate(data, *args, **kwargs)
            if isinstance(effect, torch.Tensor):
                effect = effect.detach().numpy()
            return effect

        @staticmethod
        def _f64to32(data):
            assert isinstance(data, pd.DataFrame)
            data_f64 = data.select_dtypes(include='float64')
            if len(data_f64.columns) > 0:
                data = data.copy()
                data[data.columns] = data_f64.astype('float32')
            return data

except ImportError as e:
    DeepIVWrapper = f'{e}'
    logger.warn(DeepIVWrapper)


@register()
@register(name='div')
class DeepIVFactory(BaseEstimatorFactory):
    def __init__(self, x_net=None, y_net=None, x_hidden_d=None, y_hidden_d=None, num_gaussian=5, ):
        if isinstance(DeepIVWrapper, str):
            raise ImportError(DeepIVWrapper)

        self.x_net = x_net
        self.y_net = y_net
        self.x_hidden_d = x_hidden_d
        self.y_hidden_d = y_hidden_d
        self.num_gaussian = num_gaussian

    def __call__(self, data, outcome, treatment, y_task, x_task,
                 adjustment=None, covariate=None, instrument=None, random_state=None):
        assert instrument is not None

        return DeepIVWrapper(
            x_net=self.x_net,
            y_net=self.y_net,
            x_hidden_d=self.x_hidden_d,
            y_hidden_d=self.y_hidden_d,
            num_gaussian=self.num_gaussian,
            is_discrete_outcome=y_task if isinstance(y_task, bool) else y_task != const.TASK_REGRESSION,
            is_discrete_treatment=x_task if isinstance(x_task, bool) else x_task != const.TASK_REGRESSION,
            random_state=random_state,
        )


@register()
class RLossFactory(BaseEstimatorFactory):
    def __init__(self, y_model='rf', x_model='rf'):
        self.y_model = y_model
        self.x_model = x_model

    def __call__(self, data, outcome, treatment, y_task, x_task,
                 adjustment=None, covariate=None, instrument=None, random_state=None):
        from ylearn.estimator_model.effect_score import RLoss

        return RLoss(
            y_model=self._model(data, task=y_task, estimator=self.y_model, random_state=random_state),
            x_model=self._model(data, task=x_task, estimator=self.x_model, random_state=random_state),
            cf_fold=self._cf_fold(data),
            # is_discrete_outcome=y_task != const.TASK_REGRESSION,
            is_discrete_treatment=x_task if isinstance(x_task, bool) else x_task != const.TASK_REGRESSION,
            random_state=random_state,
        )



"""
Beside effect score which measures the ability of estimating the causal
effect, we should also implement training_score measuring
performances of machine learning models.
"""
# from collections import defaultdict

import inspect
from asyncio.log import logger
import numpy as np
import pandas as pd
# from sklearn import clone
# from sklearn.preprocessing import OneHotEncoder
from sklearn.model_selection import KFold

from .double_ml import DML4CATE
from .utils import convert2array, get_wv, get_tr_ctrl
from ylearn.utils import logging

logger = logging.get_logger(__name__)

class RLoss(DML4CATE):
    """
    Estimator models for estimating causal effects can not be easily evaluated
    dut to the fact that the true effects are not directly observed. This differs
    from the usual machine learning tasks.
    
    Authors in [1] proposed a framework, a schema suggested by [2], to evaluate causal
    effects estimated by different estimator models. Roughly speaking, this
    framework is a direct application of the double machine learning methods.
    Specifically, for a causal effect model ce_model (trained in a training set)
    that is waited to be evaluated, we 
        1. Train a model y_model to estimate the outcome and a x_model to
            estimate the treatment in a validation set, which is usually not
            the same as the training set;
        2. In the validation set v, let y_res and x_res denote the differences
                y_res = y - y_pred(v),
                x_res = x - x_pred(v),
            and
                tau(v)
            denote the causal effects estimated by the ce_model in the validation
            set, then the metric of the causal effect of the ce_model is
            calculated as
                E_{V}[(y_res - x_res * tau)^2].

    Attributes
    ----------
    _is_fitted : bool
        True if the model is fitted ortherwise False.
    
    x_model : estimator
        Machine learning models for fitting x. Any such models should implement
        the fit and predict (also predict_proba if x is discrete) methods
    
    y_model : estimator
        Machine learning models for fitting y.
   
    yx_model : estimator
        Machine learning models for fitting the residual of y on residual of x.
        Currently this should be a linear regression model.
    
    adjustment_transformer : transformer
        Transformer for adjustment variables, by default None.
    
    covariate_transformer : transformer
        Transformer for covariate variables, by default None.
   
    is_discrete_treatment : bool
   
    categories : str or list
   
    random_state : int
   
    cf_fold : int
        The number of folds for performing cross fit, by default 1
    
    x_hat_dict : defaultdict(list)
        Cached values when fitting the treatment model.
   
    y_hat_dict : defaultdict(list)
        Cached values when fitting the outcome model.
    
    Methods
    ----------
    fit(data, outcome, treatment, adjustment, covariate)
        Fit the RLoss estimator model in the validation set.
   
    comp_transformer(x, categories='auto')
        Transform the discrete treatment into one-hot vectors.
    
    score(self, test_estimator)
        Calculate the metric value of the RLoss for the test_estimator in the 
        validation set.
    
    Example
    ----------
        from sklearn.ensemble import RandomForestRegressor

        from ylearn.exp_dataset.exp_data import single_binary_treatment
        from ylearn.estimator_model.meta_learner import TLearner
        
        train, val, te = single_binary_treatment()
        rloss = RLoss(
            x_model=RandomForestClassifier(),
            y_model=RandomForestRegressor(),
            cf_fold=1,
            is_discrete_treatment=True
        )
        rloss.fit(
            data=val,
            outcome=outcome,
            treatment=treatment,
            adjustment=adjustment,
            covariate=covariate,
        )
        
        est = TLearner(model=RandomForestRegressor())
        est.fit(
            data=train,
            treatment=treatment,
            outcome=outcome,
            adjustment=adjustment,
            covariate=covariate,
        )
    
    >>> rloss.score(est)
    0.20451977

    Reference
    ----------
    [1] A. Schuler, et al. A comparison of methods for model selection when
    estimating individual treatment effects. arXiv:1804.05146.
    [2] X. Nie, et al. Quasi-Oracle estimation of heterogeneous treatment effects.
        arXiv: 1712.04912.
    """
    def __init__(
        self,
        x_model,
        y_model,
        cf_fold=1,
        adjustment_transformer=None,
        covariate_transformer=None,
        random_state=2022,
        is_discrete_treatment=False,
        categories='auto',
    ):
        """
        Parameters
        ----------
        x_model : estimator
            Machine learning models for fitting x. Any such models should implement
            the fit and predict (also predict_proba if x is discrete) methods
        
        y_model : estimator
            Machine learning models for fitting y.
      
        yx_model : estimator, optional
            Machine learning models for fitting the residual of y on residual of x.
        
        cf_fold : int, optional
            The number of folds for performing cross fit, by default 1
        
        adjustment_transformer : transformer, optional
            Transformer for adjustment variables, by default None
        
        covariate_transformer : transformer, optional
            Transformer for covariate variables, by default None
       
        random_state : int, optional
            Random seed, by default 2022
       
        is_discrete_treatment : bool, optional
            If the treatment variables are discrete, set this as True, by default False
      
        categories : str, optional
        """
        self.yx_model = None

        super().__init__(
            x_model=x_model,
            y_model=y_model,
            cf_fold=cf_fold,
            adjustment_transformer=adjustment_transformer,
            covariate_transformer=covariate_transformer,
            random_state=random_state,
            is_discrete_treatment=is_discrete_treatment,
            categories=categories,
        )

    def fit(
        self,
        data,
        outcome,
        treatment,
        adjustment=None,
        covariate=None,
        combined_treatment=True,
    ):
        """Fit the RLoss estimator model. Note that the trainig of a DML has two stages, where we implement them in 
        :py:func:`_fit_1st_stage` and :py:func:`_fit_2nd_stage`.

        Parameters
        ----------
        data : pandas.DataFrame
            The dataset used for training the model
       
        outcome : str or list of str, optional
            Names of the outcome variables
       
        treatment : str or list of str
            Names of the treatment variables
        
        adjustment : str or list of str, optional
            Names of the adjustment variables, by default None
        
        covariate : str or list of str, optional
            Names of the covariate variables, by default None
        
        combined_treatment : bool
            When combined_treatment is set to True, then if there are multiple
            treatments, we can use the combined_treatment technique to covert
            the multiple discrete classification tasks into a single discrete
            classification task. For an example, if there are two different
            binary treatments:
                treatment_1: x_1 | x_1 \in {'sleep', 'run'},
                treatment_2: x_2 | x_2 \in {'study', 'work'},
            then we can convert to these two binary classification tasks into
            a single classification with 4 different classes:
                treatment: x | x \in {0, 1, 2, 3},
            where, for example, 1 stands for ('sleep' and 'stuy').

        Returns
        ----------
        instance of RLoss
            The fitted RLoss model for evaluating other estimator models in
            the validation set.
        """
        assert covariate is not None, \
            'Need covariates to use RLoss.'

        self.combined_treatment = combined_treatment

        # super().fit(
        #     data, outcome, treatment,
        #     adjustment=adjustment,
        #     covariate=covariate,
        # )
        self.outcome = outcome
        self.treatment = treatment
        self.adjustment = adjustment
        self.covariate = covariate

        y, x, w, v = convert2array(
            data, outcome, treatment, adjustment, covariate
        )
        self._w = w
        self._v = v
        self._y_d = y.shape[1]
        cfold = self.cf_fold

        if self.adjustment_transformer is not None and w is not None:
            w = self.adjustment_transformer.fit_transform(w)

        if self.covariate_transformer is not None and v is not None:
            v = self.covariate_transformer.fit_transform(v)

        if self.is_discrete_treatment:
            if self.categories == 'auto' or self.categories is None:
                categories = 'auto'
            else:
                categories = list(self.categories)

            # convert discrete treatment features to onehot vectors
            x = self.comp_transormer(x, categories)

        self._x_d = x.shape[1]
        wv = get_wv(w, v)

        # step 1: split the data
        if cfold > 1:
            cfold = int(cfold)
            folds = [
                KFold(n_splits=cfold).split(x), KFold(n_splits=cfold).split(y)
            ]
        else:
            folds = None

        # step 2: cross fit to give the estimated y and x
        self.x_hat_dict, self.y_hat_dict = super()._fit_1st_stage(
            self.x_model, self.y_model, y, x, wv, folds=folds
        )
        x_hat = self.x_hat_dict['paras'][0].reshape((x.shape))
        y_hat = self.y_hat_dict['paras'][0].reshape((y.shape))

        x_prime = x - x_hat
        y_prime = y - y_hat

        self.x_hat_dict['res'].append(x_prime)
        self.y_hat_dict['res'].append(y_prime)

        self._is_fitted = True

        return self

    def score(self, test_estimator, treat=None, control=None):
        """Calculate the RLoss as a metric of the causal effect estimated by
        the test_estimator.

        Parameters
        ----------
        test_estimator : BaseEstModel
            Any fitted estimator model for causal effect.
        
        treat : float or ndarray, optional
            In the case of single discrete treatment, treat should be an int or
            str of one of all possible treatment values which indicates the
            value of the intended treatment;
            in the case of multiple discrete treatment, treat should be a list
            or an ndarray where treat[i] indicates the value of the i-th intended
            treatment, for example, when there are multiple
            discrete treatments, array(['run', 'read']) means the treat value of
            the first treatment is taken as 'run' and that of the second treatment
            is taken as 'read';
            in the case of continuous treatment, treat should be a float or a
            ndarray, by default None
      
        control : float or ndarray, optional
            This is similar to the cases of treat, by default None

        Returns
        -------
        float
            The score for the test_estimator
        """
        x_prime, y_prime = self.x_hat_dict['res'][0], self.y_hat_dict['res'][0]
        v = self._v
        w = self._w

        data_dict = {}
        if self.covariate is not None:
            for i, name in enumerate(test_estimator.covariate):
                data_dict[name] = v[:, i]
        if self.adjustment is not None:
            for i, name in enumerate(test_estimator.adjustment):
                data_dict[name] = w[:, i]

        test_data = pd.DataFrame(data_dict)

        # shape (n, y_d, x_d)
        est_params = inspect.signature(test_estimator.estimate).parameters.keys()
        est_options = {}
        dis_tr = test_estimator.is_discrete_treatment

        if 'treat' in est_params and treat is not None:
            if hasattr(test_estimator, 'comp_transformer'):
                treat = get_tr_ctrl(treat,
                    self.comp_transormer,
                    treat=True,
                    one_hot=False,
                    discrete_treat=dis_tr,
                )
            est_options['treat'] = treat
        if 'control' in est_params and control is not None:
            if hasattr(test_estimator, 'comp_transformer'):
                control = get_tr_ctrl(control,
                    self.comp_transormer,
                    treat=False,
                    one_hot=False,
                    discrete_treat=dis_tr,
                )
            est_options['control'] = control
        test_effect = test_estimator.estimate(data=test_data, **est_options)
        logger.info(
            f'Calculating the score: {test_estimator.__repr__()} finished estimating.'
        )
        if self.is_discrete_treatment and self.combined_treatment:
            logger.info(
                'using combined treat technique for discrete treatment.'
            )
            x_prime = x_prime[:, test_estimator.treat].reshape(-1, 1)
            test_effect = test_effect.reshape(v.shape[0], self._y_d, 1)

        y_pred = np.einsum('nij, nj->ni', test_effect, x_prime)
        rloss = np.mean((y_prime - y_pred)**2, axis=0)

        return rloss

    def _fit_1st_stage(
        self,
        x_model,
        y_model,
        y, x, wv,
        folds=None,
        **kwargs
    ):
        """Fit the models in the first stage.

        Parameters
        ----------
        x_model : estimator
            Any x_model should have the fit and predict (also predict_proba if
            x is discrete) methods.
       
        y_model : estimator
            Any y_model should have the fit and predict (also predict_proba if
            y is discrete) methods.
        
        y : ndarray, shape (n, y_d)
            The outcome vector
        
        x : ndarray, shape (n, x_d)
            The treatment vector
       
        wv : ndarray, shape (n, w_d + v_d)
            The covariate and adjustment vector
        
        folds : sklearn.model_selection.KFold, optional

        Returns
        -------
        tuple of dict
            The first dict containes the values of the fitted y_model and the 
            second dict the fitted x_model
        """
        if self._x_d == 1:
            x = np.ravel(x)
        if self._y_d == 1:
            y = np.ravel(y)

        if folds is not None:
            x_folds, y_folds = folds
        else:
            x_folds, y_folds = None, None

        logger.info(
            f'_fit_1st_stage: fitting x_model {type(x_model).__name__}'
        )
        x_hat_dict = self._cross_fit(
            x_model, wv, target=x, folds=x_folds, is_ymodel=False, **kwargs
        )

        logger.info(
            f'_fit_1st_stage: fitting y_model {type(y_model).__name__}'
        )
        y_hat_dict = self._cross_fit(
            y_model, wv, target=y, folds=y_folds, is_ymodel=True, **kwargs
        )
        return (x_hat_dict, y_hat_dict)

    def __repr__(self) -> str:
        return f'RLoss'

class PredLoss:
    def __init__(self) -> None:
        pass

    def fit(self,):
        pass

    def score(self):
        pass


# class Score:
#     def __init__(self):
#         pass

#     def fit(self, data):
#         pass

#     def score(self,):
#         pass



# TODO: add copyright of the implementation of sklearn tree
import numbers
import pandas

from math import ceil
from copy import deepcopy

import numpy as np
# import pydotplus
# from ctypes import memset, sizeof, c_double, memmove

import sklearn

from sklearn.utils import check_random_state
from sklearn.utils import check_scalar

from ylearn.sklearn_ex.cloned.tree._splitter import Splitter
from ylearn.sklearn_ex.cloned.tree._splitter import BestSplitter
from ylearn.sklearn_ex.cloned.tree._splitter import RandomSplitter
from ylearn.sklearn_ex.cloned.tree._criterion import Criterion
# from ylearn.sklearn_ex.cloned.tree._criterion import MSE
from ylearn.sklearn_ex.cloned.tree._tree import Tree
from ylearn.sklearn_ex.cloned.tree._tree import DepthFirstTreeBuilder
from ylearn.sklearn_ex.cloned.tree._tree import BestFirstTreeBuilder
from sklearn.tree import plot_tree

from ..utils import logging
from ..utils._common import convert2array

from ._tree.tree_criterion import PRegCriteria
from ._tree.tree_criterion import PRegCriteria1
from ..estimator_model._tree.tree_criterion import MSE

logger = logging.get_logger(__name__)


CRITERIA = {
    "policy_reg": PRegCriteria,
    'policy_test': MSE,
    'policy_test1': PRegCriteria1
}

SPLITTERS = {
    "best": BestSplitter,
    "random": RandomSplitter,
}


class PolicyTree:
    """
    A class for finding the optimal policy for maxmizing the causal effect
    with the tree model.

    The criterion for training the tree is (in the Einstein notation)
        S = \sum_i g_{ik} y^k_{i},
    where g_{ik} = \phi(v_i)_k is a map from the covariates, v_i, to a basis vector
    which has only one nonzero element in the R^k space.

    Attributes
    ----------
    criterion : {'policy_reg'}, default to 'policy_reg' # TODO: may add more criterion

    feature_importances_ : ndarray of shape (n_features,)
        The feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the
        (normalized) total reduction of the criterion brought by that feature.

    tree_ : Tree instance
        The underlying Tree object. Please refer to
        ``help(sklearn.tree._tree.Tree)`` for attributes of Tree object and
        :ref:`sphx_glr_auto_examples_tree_plot_unveil_tree_structure.py`
        for basic usage of these attributes.

    n_outputs_ : int
        The number of outputs when fit() is performed.

    max_features_ : int
        The inferred value of max_features.

    n_features_in_ : int
        Number of features seen during fit().

    max_depth : int, default to None

    min_samples_split : int or float, default to 2

    min_samples_leaf : int or float, default to 1

    random_state : int

    max_leaf_nodes : int, default to None

    min_impurity_decrease : float, default to 0.0

    ccp_alpha : non-negative float, default to 0.0

    Methods
    ----------
    fit(data, outcome, treatment,
        adjustment=None, covariate=None, treat=None, control=None,)
        Fit the model on data.

    estimate(data=None, quantity=None)
        Estimate the value of the optimal policy for the causal effects of the treatment
        on the outcome in the data, i.e., return the value of the causal effects
        when taking the optimal treatment.    

    predict_ind(data=None,)
        Estimate the optimal policy for the causal effects of the treatment
        on the outcome in the data, i.e., return the index of the optimal treatment.

    apply(X)
        Return the index of the leaf that each sample is predicted as.

    decision_path(X)
        Return the decision path.

    _prepare4est(data)
        Prepare for the estimation of the causal effect.

    Reference
    ----------
    This implementation is based on the implementation of BaseDecisionTree
    of sklearn.
    """

    def __init__(
        self, *,
        criterion='policy_reg',
        splitter='best',
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        random_state=2022,
        max_leaf_nodes=None,
        max_features=None,
        min_impurity_decrease=0.0,
        ccp_alpha=0.0,
        min_weight_fraction_leaf=0.0,
    ):
        """
        Many parameters are similar to those of BaseDecisionTree of sklearn.

        Parameters
        ----------
        criterion : {'policy_reg'}, default to 'policy_reg' # TODO: may add more criterion
            The function to measure the quality of a split. The criterion for
            training the tree is (in the Einstein notation)
                    S = \sum_i g_{ik} y^k_{i},
            where g_{ik} = \phi(v_i)_k is a map from the covariates, v_i, to a
            basis vector which has only one nonzero element in the R^k space. By
            using this criterion, the aim of the model is to find the index of the
            treatment which will render the max causal effect, i.e., finding the
            optimal policy.

        splitter : {"best", "random"}, default="best"
            The strategy used to choose the split at each node. Supported
            strategies are "best" to choose the best split and "random" to choose
            the best random split.

        max_depth : int, default=None
            The maximum depth of the tree. If None, then nodes are expanded until
            all leaves are pure or until all leaves contain less than
            min_samples_split samples.

        min_samples_split : int or float, default=2
            The minimum number of samples required to split an internal node:
            - If int, then consider `min_samples_split` as the minimum number.
            - If float, then `min_samples_split` is a fraction and
            `ceil(min_samples_split * n_samples)` are the minimum
            number of samples for each split.

        min_samples_leaf : int or float, default=1
            The minimum number of samples required to be at a leaf node.
            A split point at any depth will only be considered if it leaves at
            least ``min_samples_leaf`` training samples in each of the left and
            right branches.  This may have the effect of smoothing the model,
            especially in regression.
            - If int, then consider `min_samples_leaf` as the minimum number.
            - If float, then `min_samples_leaf` is a fraction and
            `ceil(min_samples_leaf * n_samples)` are the minimum
            number of samples for each node.

        min_weight_fraction_leaf : float, default=0.0
            The minimum weighted fraction of the sum total of weights (of all
            the input samples) required to be at a leaf node. Samples have
            equal weight when sample_weight is not provided.

        max_features : int, float or {"sqrt", "log2"}, default=None
            The number of features to consider when looking for the best split:
            - If int, then consider `max_features` features at each split.
            - If float, then `max_features` is a fraction and
            `int(max_features * n_features)` features are considered at each
            split.
            - If "sqrt", then `max_features=sqrt(n_features)`.
            - If "log2", then `max_features=log2(n_features)`.
            - If None, then `max_features=n_features`.

        random_state : int
            Controls the randomness of the estimator.

        max_leaf_nodes : int, default to None
            Grow a tree with ``max_leaf_nodes`` in best-first fashion.
            Best nodes are defined as relative reduction in impurity.
            If None then unlimited number of leaf nodes.

        min_impurity_decrease : float, default=0.0
            A node will be split if this split induces a decrease of the impurity
            greater than or equal to this value.
            The weighted impurity decrease equation is the following::
                N_t / N * (impurity - N_t_R / N_t * right_impurity
                                    - N_t_L / N_t * left_impurity)
            where ``N`` is the total number of samples, ``N_t`` is the number of
            samples at the current node, ``N_t_L`` is the number of samples in the
            left child, and ``N_t_R`` is the number of samples in the right child.
            ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
            if ``sample_weight`` is passed.

        ccp_alpha : non-negative float, default to 0.0
            Value for pruning the tree. #TODO: not implemented yet.
        """
        self._is_fitted = False

        self.criterion = criterion
        self.splitter = splitter
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.ccp_alpha = ccp_alpha
        self.random_state = random_state
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features

    def fit(
        self,
        data,
        covariate, *,
        effect=None,
        effect_array=None,
        est_model=None,
        sample_weight=None,
        **kwargs,
    ):
        # TODO: consider possibility for generalizing continuous treatment
        """Fit the model on data.

        Parameters
        ----------
        data : pandas.DataFrame

        covariate : str or list of str
            Names of the covariates.

        effect : str or list of str
            Names of the causal effects vectors.

        effect_array : ndarray of shape (n, n_treatments)
            The causal effect that waited to be fitted by  :class:`PolicyTree`. If this is not provided and `est_model` is None, then `effect` can not be None.
        
        est_model : Any valid estimator model in ylearn
            If `effect=None` and `effect_array=None`, then `est_model` can not be None and the causal
            effect will be estimated by the `est_model`.
        
        Returns
        ----------
        instance of self
        """
        ce, v = convert2array(data, effect, covariate)
        self.covariate = covariate

        if effect_array is not None:
            ce = effect_array
        else:
            if effect is None:
                assert est_model is not None, \
                    'The causal effect is not provided and no estimator model is'
                'provided to estimate it from the data.'

                assert est_model._is_fitted

                # assert est_model.covariate == covariate
                self.covariate = est_model.covariate
                v = convert2array(data, self.covariate)[0]

                if hasattr(est_model, 'covariate_transformer') and est_model.covariate_transformer is not None:
                    self.cov_transformer = est_model.covariate_transformer
                    v = self.cov_transformer.transform(v)

                ce = est_model.effect_nji(data)
                ce = self._check_ce(ce)

        self._v = v

        # check random state
        random_state = check_random_state(self.random_state)

        # Determine output settings
        n_samples, self.n_features_in_ = v.shape  # dimension of the input

        # reshape ce if necessary
        if ce.ndim == 1:
            ce = ce.reshape(-1, 1)

        self.n_outputs_ = ce.shape[1]

        # Check parameters
        if self.max_depth is not None:
            check_scalar(
                self.max_depth,
                name='max_depth',
                target_type=numbers.Integral,
                min_val=1,
            )
        max_depth = np.iinfo(np.int32).max if self.max_depth is None \
            else self.max_depth

        # check self.min_samples_leaf
        if isinstance(self.min_samples_leaf, numbers.Integral):
            check_scalar(
                self.min_samples_leaf,
                name='min_samples_leaf',
                target_type=numbers.Integral,
                min_val=1,
            )
            min_samples_leaf = self.min_samples_leaf
        else:
            check_scalar(
                self.min_samples_leaf,
                name='min_samples_leaf',
                target_type=numbers.Real,
                min_val=0.0,
                include_boundaries='neither',
            )
            min_samples_leaf = int(ceil(self.min_samples_leaf * n_samples))

        # check self.min_samples_split
        if isinstance(self.min_samples_split, numbers.Integral):
            check_scalar(
                self.min_samples_split,
                name='min_samples_split',
                target_type=numbers.Integral,
                min_val=2,
            )
            min_samples_split = self.min_samples_split
        else:
            check_scalar(
                self.min_samples_split,
                name='min_samples_split',
                target_type=numbers.Real,
                min_val=0.0,
                max_val=1.0,
                include_boundaries='right',
            )
            min_samples_split = int(ceil(self.min_samples_split * n_samples))
            min_samples_split = max(2, min_samples_split)

        min_samples_split = max(min_samples_split, 2 * min_samples_leaf)

        # check self.max_leaf_nodes
        if self.max_leaf_nodes is not None:
            check_scalar(
                self.max_leaf_nodes,
                name='max_leaf_nodes',
                target_type=numbers.Integral,
                min_val=2,
            )

        max_leaf_nodes = -1 if self.max_leaf_nodes is None else self.max_leaf_nodes

        # check min_weight_fraction_leaf
        check_scalar(
            self.min_weight_fraction_leaf,
            name='min_weight_fraction_leaf',
            target_type=numbers.Real,
            min_val=0.0,
            max_val=0.5,
        )

        # check max_features
        if isinstance(self.max_features, str):
            if self.max_features == 'sqrt':
                max_features = max(1, int(np.sqrt(self.n_features_in_)))
            elif self.max_features == 'log2':
                max_features = max(1, int(np.log2(self.n_features_in_)))
            else:
                raise ValueError(
                    'Invalid value for max_features. Allowed string values'
                    f'Allowed string values are "sqrt" or "log2", but was given {self.max_features}.'
                )
        elif self.max_features is None:
            max_features = self.n_features_in_
        elif isinstance(self.max_features, numbers.Integral):
            check_scalar(
                self.max_features,
                name='max_features',
                target_type=numbers.Integral,
                min_val=1,
                include_boundaries='left',
            )
            max_features = self.max_features
        else:
            check_scalar(
                self.max_features,
                name='max_features',
                target_type=numbers.Real,
                min_val=0.0,
                max_val=1.0,
                include_boundaries='right',
            )
            if self.max_features > 0.0:
                max_features = max(
                    1, int(self.max_features * self.n_features_in_)
                )
            else:
                max_features = 0

        self.max_features_ = max_features

        check_scalar(
            self.min_impurity_decrease,
            name='min_impurity_decrease',
            target_type=numbers.Real,
            min_val=0.0,
        )

        if len(ce) != n_samples:
            raise ValueError(
                f'The number of labels {len(ce)} does not match the number of samples'
            )

        # set min_weight_leaf
        if sample_weight is None:
            min_weight_leaf = self.min_weight_fraction_leaf * n_samples
        else:
            min_weight_leaf = self.min_weight_fraction_leaf * \
                np.sum(sample_weight)

        # Build tree step 1. Set up criterion
        criterion = self.criterion

        if not isinstance(criterion, Criterion):
            # assert criterion == 'policy_clf' or criterion == 'policy_reg', f'The str criterion should be policy, but was given {self.criterion}'
            criterion = CRITERIA[self.criterion](self.n_outputs_, n_samples)
        else:
            criterion = deepcopy(self.criterion)

        logger.info(
            f'Start building the policy tree with criterion {type(criterion).__name__}'
        )

        # Build tree step 2. Define splitter
        splitter = self.splitter

        if not isinstance(self.splitter, Splitter):
            splitter = SPLITTERS[self.splitter](
                criterion,
                self.max_features_,
                min_samples_leaf,
                min_weight_leaf,
                random_state,
            )

        logger.info(
            f'Building the policy tree with splitter {type(splitter).__name__}'
        )

        # Build tree step 3. Define the tree
        self.tree_ = Tree(
            self.n_features_in_,
            np.array([1] * self.n_outputs_, dtype=np.intp),
            self.n_outputs_,
        )

        # Build tree step 3. Build the tree
        if max_leaf_nodes < 0:
            builder = DepthFirstTreeBuilder(
                    splitter,
                    min_samples_split,
                    min_samples_leaf,
                    min_weight_leaf,
                    max_depth,
                    self.min_impurity_decrease,
                )
        else:
            builder = BestFirstTreeBuilder(
                splitter,
                min_samples_split,
                min_samples_leaf,
                min_weight_leaf,
                max_depth,
                max_leaf_nodes,
                self.min_impurity_decrease,
            )

        logger.info(
            f'Building the policy tree with builder {type(builder).__name__}'
        )

        builder.build(self.tree_, v, ce, sample_weight)

        self._is_fitted = True

        return self

    def predict_ind(self, data=None):
        """Estimate the optimal policy for the causal effects of the treatment
        on the outcome in the data, i.e., return the index of the optimal treatment.

        Parameters
        ----------
        data : pandas.DataFrame, optional. Defaults to None
            If None, data will be set as the training data.    

        Returns
        -------
        ndarray or float, optional
            The index of the optimal treatment dimension. 
        """
        pred = self._prepare4est(data).argmax(1)

        return pred

    def predict_opt_effect(self, data=None):
        """Estimate the value of the optimal policy for the causal effects of the treatment
        on the outcome in the data, i.e., return the value of the causal effects
        when taking the optimal treatment.

        Parameters
        ----------
        data : pandas.DataFrame, optional. Defaults to None
            If None, data will be set as the training data.    

        Returns
        -------
        ndarray or float, optional
            The estimated causal effect with the optimal treatment value.
        """
        pred = self._prepare4est(data)

        return pred

    def _prepare4est(self, data=None):
        assert self._is_fitted, 'The model is not fitted yet.'

        v = self._check_features(v=None, data=data)

        proba = self.tree_.predict(v)
        n_samples = v.shape[0]

        if self.criterion == 'policy_reg' or self.criterion == 'policy_test1':
            if self.n_outputs_ == 1:
                return proba[:, 0]
            else:
                return proba[:, :, 0]
        else:
            pass

    def _prune_tree(self):
        raise NotImplemented()

    def _check_ce(self, ce):
        if ce.ndim == 1:
            ce = ce.reshape(-1, 1)
        else:
            n, _yx_d = ce.shape[0], ce.shape[1]
            if ce.ndim == 3:
                assert _yx_d == 1, f'Expect effect array with shape {(n, ce.shape[2])}, but was given {(n, _yx_d, ce.shape[2])}'

                ce = ce.reshape(n, -1)
            elif ce.ndim == 2:
                pass
            else:
                raise ValueError(
                    f'Expect effect array with shape (num_samples, num_treatments) but was given {ce.shape}')

        return ce

    def _check_features(self, *, v=None, data=None):
        """Validate the training data on predict (probabilities)."""

        if v is not None:
            v = v.reshape(-1, 1) if v.ndim == 1 else v
            assert v.shape[1] == self.n_features_in_
            v = v.astype(np.float32)

            return v

        if data is None:
            v = self._v
        else:
            assert isinstance(data, pandas.DataFrame)

            v = convert2array(data, self.covariate)[0]
            if hasattr(self, 'cov_transformer'):
                v = self.cov_transformer.transform(v)

            assert v.shape[1] == self.tree_.n_features

        v = v.astype(np.float32)

        return v

    def apply(self, *, v=None, data=None):
        """Return the index of the leaf that each sample is predicted as.

        Parameters
        ----------
        data : DataFrame, optional
            The input samples. The data must contains columns of the covariates
            used for training the model. If None, the training data will be
            passed as input samples , by default None

        Returns
        -------
        v_leaves : array-like of shape (n_samples, )
            For each datapoint v_i in v, return the index of the leaf v_i
            ends up in. Leaves are numbered within
            ``[0; self.tree_.node_count)``, possibly with gaps in the
            numbering.
        """
        assert self._is_fitted, 'The model is not fitted yet.'

        v = self._check_features(v=v, data=data)

        return self.tree_.apply(v)

    def decision_path(self, *, v=None, data=None):
        """Return the decision path in the tree.

        Parameters
        ----------
        data : DataFrame, optional
            The input samples. The data must contains columns of the covariates
            used for training the model. If None, the training data will be
            passed as input samples , by default None

        Returns
        -------
        indicator : sparse matrix of shape (n_samples, n_nodes)
            Return a node indicator CSR matrix where non zero elements
            indicates that the samples goes through the nodes.
        """
        assert self._is_fitted, 'The model is not fitted yet.'

        v = self._check_features(v=v, data=data)

        return self.tree_.decision_path(v)

    @property
    def feature_importance(self):
        """Return the feature importances.
        The importance of a feature is computed as the (normalized) total
        reduction of the criterion brought by that feature.
        It is also known as the Gini importance.
        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.

        Returns
        -------
        feature_importances_ : ndarray of shape (n_features,)
            Normalized total reduction of criteria by feature
            (Gini importance).
        """
        assert self._is_fitted, 'The model is not fitted yet.'

        return self.tree_.compute_feature_importances()

    def get_depth(self):
        """Return the depth of the policy tree.
        The depth of a tree is the maximum distance between the root
        and any leaf.
        Returns
        -------
        self.tree_.max_depth : int
            The maximum depth of the tree.
        """
        assert self._is_fitted
        return self.tree_.max_depth

    def get_n_leaves(self):
        """Return the number of leaves of the policy tree.
        Returns
        -------
        self.tree_.n_leaves : int
            Number of leaves.
        """
        assert self._is_fitted
        return self.tree_.n_leaves

    def plot(
        self, *,
        max_depth=None,
        feature_names=None,
        class_names=None,
        label='all',
        filled=False,
        node_ids=False,
        proportion=False,
        rounded=False,
        precision=3,
        ax=None,
        fontsize=None
    ):
        """Plot a policy tree.
        The sample counts that are shown are weighted with any sample_weights that
        might be present.
        The visualization is fit automatically to the size of the axis.
        Use the ``figsize`` or ``dpi`` arguments of ``plt.figure``  to control
        the size of the rendering.

        Parameters
        ----------        
        max_depth : int, default=None
            The maximum depth of the representation. If None, the tree is fully
            generated.

        class_names : list of str or bool, default=None
            Names of each of the target classes in ascending numerical order.
            Only relevant for classification and not supported for multi-output.
            If ``True``, shows a symbolic representation of the class name.

        label : {'all', 'root', 'none'}, default='all'
            Whether to show informative labels for impurity, etc.
            Options include 'all' to show at every node, 'root' to show only at
            the top root node, or 'none' to not show at any node.

        filled : bool, default=False
            When set to ``True``, paint nodes to indicate majority class for
            classification, extremity of values for regression, or purity of node
            for multi-output.

        impurity : bool, default=True
            When set to ``True``, show the impurity at each node.

        node_ids : bool, default=False
            When set to ``True``, show the ID number on each node.

        proportion : bool, default=False
            When set to ``True``, change the display of 'values' and/or 'samples'
            to be proportions and percentages respectively.

        rounded : bool, default=False
            When set to ``True``, draw node boxes with rounded corners and use
            Helvetica fonts instead of Times-Roman.

        precision : int, default=3
            Number of digits of precision for floating point in the values of
            impurity, threshold and value attributes of each node.

        ax : matplotlib axis, default=None
            Axes to plot to. If None, use current axis. Any previous content
            is cleared.

        fontsize : int, default=None
            Size of text font. If None, determined automatically to fit figure.

        Returns
        -------
        annotations : list of artists
            List containing the artists for the annotation boxes making up the
            tree.
        """
        assert self._is_fitted

        impurity = False
        feature_names = self.covariate if feature_names is None else feature_names

        return plot_tree(
            self,
            max_depth=max_depth,
            feature_names=feature_names,
            class_names=class_names,
            label=label,
            filled=filled,
            impurity=impurity,
            node_ids=node_ids,
            proportion=proportion,
            rounded=rounded,
            precision=precision,
            ax=ax,
            fontsize=fontsize,
        )

    @property
    def n_features_(self):
        return self.n_features_in_
